/******************************************************************************

 FILENAME:             VERILOG.LEX	

 CREATION DATE:        August 3, 1998

 AUTHOR:               Ben Foster

 PURPOSE:              This file performs the lexical analysis for its 
                       companion file, "Verilog.yacc".  It recognizes strings,
		       symbols, and numbers in the input Verilog file (read 
		       in by the "main" function contained in "Verilog.yacc), 
		       returning the appropriate token and value for each 
		       recognized symbol in the Verilog file.  

 SIDE EFFECTS:         None

 COMMENTS:             None

 HISTORY:              Unmodified

******************************************************************************/

%{
#include <stdio.h>
#include <string.h>
#include "y.tab.h"

#define BUFFER_SIZE 1024

void erase_underscores(void);
void get_string(void);
void get_char(void);

static char buffer[BUFFER_SIZE];

char linebuf[500];

int lineno = 0;

%}

%p 5000         /*positions*/
%e 1500         /*nodes*/
%a 1500         /*packed transitions*/
%n 1000         /*???*/
%o 1500         /*output slots*/

         /*beginning of lexical analysis, starting with Verilog keywords*/
%%

"always"           { return ALWAYS; }
"and"              { return AND; }
"assign"           { return ASSIGN; }
"begin"            { return BEGN; }
"body"             { return BODY; }
"buf"              { return BUF; }
"bufif0"           { return BUFIF0; }
"bufif1"           { return BUFIF1; }
"case"             { return CASE; }
"casex"            { return CASEX; }
"casez"            { return CASEZ; }
"cmos"             { return CMOS; }
"deassign"         { return DEASSIGN; }
"default"          { return DEFAULT; }
"defparam"         { return DEFPARAM; }
"dff"              { return DFF; }
"disable"          { return DISABLE; }
"edge"             { return EDGE; }
"else"             { return ELSE; }
"end"              { return END; }
"endcase"          { return ENDCASE; }
"endfunction"      { return ENDFUNCTION; }
"endmodule"        { return ENDMODULE; }
"endprimitive"     { return ENDPRIMITIVE; }
"endspecify"       { return ENDSPECIFY; }
"endtable"         { return ENDTABLE; }
"endtask"          { return ENDTASK; }
"endunit"          { return ENDUNIT; }
"event"            { return EVENT; }
"for"              { return FOR; }
"force"            { return FORCE; }
"forever"          { return FOREVER; }
"fork"             { return FORK; }
"function"         { return FUNCTION; }
"highz0"           { return HIGHZ0; }
"highz1"           { return HIGHZ1; }
"hold"             { return HOLD; }
"if"               { return IF; }
"ifnone"           { return IFNONE; }
"in"               { return IN; }
"inv"              { return INV; }
"initial"          { return INITL; }
"inout"            { return INOUT; }
"input"            { return INPUT; }
"integer"          { return INTEGER; }
"join"             { return JOIN; }
"large"            { return LARGE; }
"macromodule"      { return MACROMODULE; }
"medium"           { return MEDIUM; }
"module"           { return MODULE; }
"nand"             { return NAND; }
"negedge"          { return NEGEDGE; }
"nmos"             { return NMOS; }
"nor"              { return NOR; }
"not"              { return NOT; }
"notif0"           { return NOTIF0; }
"notif1"           { return NOTIF1; }
"or"               { return OR; }
"out"              { return OUT; }
"output"           { return OUTPUT; }
"PATHPULSE"        { return PATHPULSE; }
"period"           { return PERIOD; }
"pmos"             { return PMOS; }
"posedge"          { return POSEDGE; }
"primitive"        { return PRIMITIVE; }
"pull0"            { return PULL0; }
"pull1"            { return PULL1; }
"pulldown"         { return PULLDOWN; }
"pullup"           { return PULLUP; }
"rcmos"            { return RCMOS; }
"reg"              { return REG; }
"recovery"         { return RECOVERY; }
"release"          { return RELEASE; }
"repeat"           { return REPEAT; }
"rnmos"            { return RNMOS; }
"rpmos"            { return RPMOS; }
"rtran"            { return RTRAN; }
"rtranif0"         { return RTRANIF0; }
"rtranif1"         { return RTRANIF1; }
"scalared"         { return SCALARED; }
"setup"            { return SETUP; }
"setuphold"        { return SETUPHOLD; }
"skew"             { return SKEW; }
"small"            { return SMALL; }
"specify"          { return SPECIFY; }
"specparam"        { return SPECPARAM; }
"strong0"          { return STRONG0; }
"strong1"          { return STRONG1; }
"supply0"          { return SUPPLY0; }
"supply1"          { return SUPPLY1; }
"table"            { return TABLE; }
"task"             { return TASK; }
"time"             { return TIME; }
"tran"             { return TRAN; }
"tranif0"          { return TRANIF0; }
"tranif1"          { return TRANIF1; }
"tri"              { return TRI; }
"tri0"             { return TRI0; }
"tri1"             { return TRI1; }
"triand"           { return TRIAND; }
"trireg"           { return TRIREG; }
"trior"            { return TRIOR; }
"unit"             { return UNIT; }
"vectored"         { return VECTORED; }
"wait"             { return WAIT; }
"wand"             { return WAND; }
"weak0"            { return WEAK0; }
"weak1"            { return WEAK1; }
"while"            { return WHILE; }
"width"            { return WIDTH; }
"wire"             { return WIRE; }
"wor"              { return WOR; }
"xnor"             { return XNOR; }
"xor"              { return XOR; }

"//"[^\n]*        ;

"/*"[^"*/"]*"*/"  ;

^![^\n]*          ;

[a-zA-Z][a-zA-Z0-9_$]* {
  get_string();
  return SIMPLE_IDENTIFIER;
}

\\[^ \t\n] {
  get_string();
  return ESCAPED_IDENTIFIER;
}

\"[^\n\"] {
  get_string();
  return STRING;
}

[01] {
  yylval.n = atoi(yytext); 
  return ZERO_OR_ONE;
}

1\'[bB][01] {
  get_string();
  return ONE_KNOWN_BIT;
}

1\'[bB][xX] {
  get_string();
  return UNKNOWN_BIT;
}

\'[bB][01] {
  get_string();
  return KNOWN_BIT;
}

[xX] {
  get_string();
  return EX;
}

[bB] {
  get_string();
  return LEVEL_SYMBOL;
}

[rRfFpPnN] {
  get_string();
  return EDGE_SYMBOL;
}

01 {
  get_string();
  return EDGE_DESCRIPTOR;
}

10 {
  yylval.n = atoi(yytext);
  return EDGE_OR_UNSIGNED;
}

x[01] {
  get_string();
  return EDGE_DESCRIPTOR;
}

[01]x {
  get_string();
  return EDGE_DESCRIPTOR;
}

[1-9][0-9_]* {              /*collect unsigned radix-unspecified numbers*/ 
  erase_underscores();      /*erase the underscore characters*/
  yylval.n = atoi(yylval.txt);
  return UNSIGNED_NUMBER;
}

[0-9]+\.[0-9]+ {
  erase_underscores();
  return REAL_NUMBER;
}

[0-9]*\.?[0-9]+[eE][-+]?[0-9]+ {
  erase_underscores();
  return REAL_NUMBER;
}

'[b|B][xXzZ01][xXzZ01_]* {  /*collect binary numbers */           
  erase_underscores();      /*erase the underscore characters*/
  return BINARY_NUMBER;
}

'[o|O][xXzZ0-7][xXzZ0-7_]* {/*collect octal numbers */           
  erase_underscores();      /*erase the underscore characters*/
  return OCTAL_NUMBER;
}

'[d|D][0-9][0-9_]* {        /*collect decimal numbers */           
  erase_underscores();      /*erase the underscore characters*/
  yylval.n = atoi(yylval.txt);
  return DECIMAL_NUMBER;
}

'[h|H][xXzZ0-9a-fA-F][xXzZ0-9a-fA-F_]* { /*collect hexadecimal numbers */
  erase_underscores();      /*erase the underscore characters*/
  return HEX_NUMBER;
}

"+" {
  get_char();
  return PLUS;
}

- {
  get_char();
  return MINUS;
}

[&|^] {
  get_string();
  return OPERATOR;
}

"^~" {
  get_string();
  return OPERATOR;
}

"~^" {
  get_string();
  return OPERATOR;
}

"*" {
  get_string();
  return EDGE_OR_BINARY_OPERATOR;
}

[/%<>] {
  get_string();
  return BINARY_OPERATOR;
}

[=!]= {
  get_string();
  return BINARY_OPERATOR;
}

"<=" {
  get_string();
  return LESSOREQUAL;
}

"=>" {
  get_string();
  return DOUBLERIGHTARROW;
}

[=!]== {
  get_string();
  return EQUIV_EXPR;
}

"&&" {
  get_string();
  return BINARY_OPERATOR;
}

"||" {
  get_string();
  return BINARY_OPERATOR;
}

">>" {
  get_string();
  return BINARY_OPERATOR;
}

"<<" {
  get_string();
  return BINARY_OPERATOR;
}

~[&|] {
  get_string();
  return UNARY_OPERATOR;
}

~ {
  get_string();
  return TILDA;
}

! {
  get_string();
  return UNARY_OPERATOR;
}

"&&&" {
  get_string();
  return TRIAMPERSAND;
}

= {
  get_char();
  return EQUAL;
}

"," {
  get_char();
  return COMMA;
}

"(" {
  get_char();
  return OPENPAREN;
}

")" {
  get_char();
  return CLOSEPAREN;
}

"{" {
  get_char();
  return OPENCURLYBRAC;
}

"}" {
  get_char();
  return CLOSECURLYBRAC;
}

"[" {
  get_char();
  return OPENSQBRAC;
}

"]" {
  get_char();
  return CLOSESQBRAC;
}

"#" {
  get_char();
  return POUND;
}

"@" {
  get_char();
  return AT;
}

":" {
  get_char();
  return COLON;
}

";" {
  get_char();
  return SEMICOLON;
}

"?" {
  get_char();
  return QUESTIONMARK;
}

\\ {
  get_char();
  return BACKSLASH;
}

-> {
  get_string();
  return RIGHTARROW;
}

"*>" {
  get_string();
  return STARARROWRIGHT;
}

"$" {
  get_char();
  return DOLLAR;
}

"." {
  get_char();
  return DOT;
}

:= {
  get_string();
  return DES_ASSIGNMENT;
}

"'" {
  get_char();
  return SINGLE_QUOTE;
}

.              ;
\n.* {strcpy(linebuf, yytext+1);    /*copy the next line */
 lineno++;                          /*in case of an error*/
 yyless(1);                 /*return all but '\n' for analysis*/
}

%%        /*end of lexical analysis section*/

/******************************************************************************
 FILENAME:             VERILOG.LEX
 FUNCTION NAME:        get_string(void)
 CREATION DATE:        August 3, 1998
 AUTHOR:               Ben Foster
 PARAMETERS:           No direct parameters
 PURPOSE:              To copy the value stored in "yytext" into yacc's char
                           string, "yylval.txt".
 SIDE EFFECTS:         The value of "yylval.txt" gets set to the value stored
                           in "yytext".  
 COMMENTS:             This is the standard method of associating a string 
                           value with a token 
 HISTORY:              Unmodified
******************************************************************************/

/* read in a string of indeterminate length*/
void get_string(void){
  sscanf(yytext, "%s", buffer);
  yylval.txt = (char *)malloc(strlen(buffer)+1);
  buffer[strlen(buffer)+1] = '\0';
  strcpy(yylval.txt, buffer);
}
/* END OF get_string(void) */

/******************************************************************************
 FILENAME:             VERILOG.LEX
 FUNCTION NAME:        get_char(void)
 CREATION DATE:        August 3, 1998
 AUTHOR:               Ben Foster
 PARAMETERS:           No direct parameters
 PURPOSE:              To copy the value stored in "yytext" into yacc's 
                           character variable, "yylval.c".
 SIDE EFFECTS:         The value of "yylval.c" gets set to the value stored
                           in "yytext".  
 COMMENTS:             This is the standard method of associating a char 
                           value with a token 
 HISTORY:              Unmodified
******************************************************************************/
/* read in a character*/
void get_char(void){
  yylval.c = yytext[0];
}
/* END OF get_char(void) */

/******************************************************************************
 FILENAME:             VERILOG.LEX
 FUNCTION NAME:        erase_underscores(int i)
 CREATION DATE:        August 3, 1998
 AUTHOR:               Ben Foster
 PARAMETERS:           No direct parameters
 PURPOSE:              To copy the value stored in "yytext" into yacc's 
                           char string variable, "yylval.txt", after 
			   skipping the underscore characters.
 SIDE EFFECTS:         The value of "yylval.c" gets set to the value stored
                           in "yytext", minus the underscore characters.  
 COMMENTS:             Underscore characters may be present in numbers in 
                           Verilog models, but not in MCS.  This routine 
			   simply eliminates the underscores from the 
			   numbers.
 HISTORY:              Unmodified
******************************************************************************/
/* erase underscore characters found in numbers*/
void erase_underscores(void){
  int i=0, j=0;

  while(yytext[i]!='\0'){
    if(yytext[i]!='_'){
      buffer[j] = yytext[i];
      i++;
      j++;
    }
    else i++;
  }
  buffer[j] = '\0';

  yylval.txt=(char *)malloc(strlen(buffer));
  strcpy(yylval.txt, buffer);
}
/*END OF erase_underscores(void)*/

/******************************************************************************
 FILENAME:             VERILOG.LEX
 FUNCTION NAME:        yyerror(char *s)
 CREATION DATE:        August 3, 1998
 AUTHOR:               Ben Foster
 PARAMETERS:           Char string s, automatically given to yyerror(char *s)
                           by lex when an error occurs.
 PURPOSE:              For better error documentation to aid in debugging
 SIDE EFFECTS:         None
 COMMENTS:             "yyerror(char *s) is automatically called by the
                           lexical analyser when an error is detected.

                       "linebuf" is altered with every newline character
                           encountered.  "linebuf" contains a copy of each 
			   line during the time that it is being scanned by 
			   lex (see end of lexical analysis section above).
 HISTORY:              Unmodified
******************************************************************************/
void yyerror(char *s){
  printf("%d: %s at %s in this line:\n%s\n",
	 lineno+1, s, yytext, linebuf);
}
/*END OF yyerror(char *s)*/
