/*******************************************************************************************************************

 FILENAME:		VERILOG.YACC

 CREATION DATE:		August 3, 1998
	
 AUTHORS:		Ben Foster,
			Prof. Alva L. Couch

 PURPOSE:		This program accepts an input file written in the Verilog Hardware Description Language,
			parses this file, and using the data collected from this file creates a DEE netlist file
			which is used by the MCS Concurrent Fault Simulator. 

			This file requires its companion lex file, "verilog.lex", for lexical analysis of the 
			input Verilog file. 

 ALGORITHM:		This parser begins by first recognizing a Verilog module declaration, denoted by the 
			keyword "module".  Once it is inside the module, the parser searches for inputs, outputs,
			bidirectional wires, internal wires, and logic gates.  

			Two major data structures handle all the information scanned in by the program.  "Elemlist"
			stores all information related to actual gates and all physical wires. "Wirelist" acts as a 
			list of the links connecting all the elements stored in Elemlist.  Wirelist is in essence 
			a representation of all the hidden internal wires.  Both lists are connected to each other 
			by connection lists, which represent the fanin connections for each element and the fanout 
			connections for each internal wire.  Each element in Elemlist and Wirelist has its own 
			connection list.  Elemlist's connection lists point to the internal wires in Wirelist which
			fan in to Elemlist's current element.  Wirelist's connection lists point to the elements in
			Elemlist which the current wire fans out to.  Wirelist has an additional connection, a 
			pointer to an element, which stores the element from which the current wire originates 
			(the stem).  Elemlist's elements cannot have a similar structure which points to its
			output wire, since one or more wires may originate from an element as output (in the 
			case of behavioral elements).

			Inputs, outputs, and bidirectional wires are treated as actual wires.  Each are assigned 
			an element in the elemlist.  Their connections are left empty until their names are found 
			again in the logic gate description.
			
			Wires which are labeled as wires and have no delay are ignored.

			A new element is created for each logic gate read in by the parser.  After the element is 
			created and a few basic parameters are set, the element is linked to the end of the 
			element list.  The next step is the creation of the element's connection list and the 
			element's output's connection list.  These connection lists are determined from each 
			logic gate's list of terminal identifiers, which are the identifiers enclosed in 
			parentheses and separated by commas in the Verilog description file.  In all cases except
		 	for D-type flip flops, the first identifier in the list is the output terminal of the logic
			gate.  
				When an output terminal is encountered, neither the element's nor the wire's 
			connection list is altered.  The only link created is the wire's pointer to the element 
			which drives that particular internal wire.  The program searches through the wirelist, 
			looking for a match in identifier names.  (This is the biggest time useage the program 
			experiences.  The time could be cut significantly by implementing an array of wirelists
			indexed by a hash table.)  When a match of the identifiers is found, the loop is broken and 
			the wirelist pointer points to the desired wire.  If no match is found, a new wire is 
			created and added to the tail of the wirelist.  In either case, the wirelist pointer will
			point to the desired wire.  This wire's element pointer is updated to point to the current
			element.
				When an input terminal is encountered, both the wire's and the element's connection
			lists are updated.  The same method of list searching explained above is used to find the 
			wire which matches the current input identifier.  As before, if the wire is not found in the
			wirelist, a new wire is created and added to the list.  To create the wire's connection 
			list, a block of memory is allocated for the wire's connection.  That connection is then 
			added to the tail of its connection list.  Finally, the connection's element pointer (ept)
			is set to point at the current element in the elemlist.  To create the element's connection
			list, a block of memory is allocated for the element's connection.  That connection is then
			added to the tail of its connection list.  To finish the current connection, the 
			connection's wire pointer (wpt) is set to point at the current wire in the wirelist. 
			(See Figure 1). 
				Both of these procedures are repeated until the current element's data are read in,
			then the parser will move to the next element in the Verilog description file. 


							w1   ___        	    
							____[   \  w4     
							    | G2 )---[]         
					w2   ___	    [___/      
					____[   \       
					____| G1 )         
				        w3  [___/      

				Current line in Verilog description file:
				
					and G2(w4, w1, w5);   (wire w4 is the output, wires w1 and w5 are inputs)
						      
						Figure 1      (NOT COMPLETED)

			Internal Wires with delay are handled slightly differently.  A delayed wire is inserted 
			into the elemlist as an element with a delay.  The internal wire is added to the wirelist 
			as well, with its element pointer pointed to its delayed wire element in the elemlist.  
			When this wire is referenced in the Verilog description file as an input terminal to a 
			physical gate, the normal actions of connecting the wire's and the element's connection 
			lists take place.  The main difference is seen when the wire appears as an output terminal
			of an element in the description file.  To make a proper connection, a new wire must be 
			added to the wirelist to connect the current element and the delayed wire (See Fig. 2).

					   physical    conn.    delayed   internal
				  	     gate      point      wire      wire
					                 |         |         |     
					     ___         |         |         | 	   ____
					____[   \        V         V         V	  |     
					____|    )-[]         []~~~~~~~[]---------|     =>     other 
					    [___/                                 |____     connections

							Figure 2

			(Normally, the gate would be connected to the internal wire, but this would have the effect
			of bypassing the delayed wire.)  The program first checks to see if the gate pointed to by
			the internal wire is a delayed wire.  If so, a new wire is created and linked to the 
			end of the wirelist.  
			
					wirelist = MALLOC_TYPE(struct Wire);		
		
			It is NOT given a name; it is treated as an invisible wire.  
			
					wirelist->name = NULL;
		
			The new wire's element pointer is set to point at the physical gate.  
			
					wirelist->elem = elemlist;   (elemlist points to the physical gate)
		
			The new wire's connection list's element pointer is set to point at the delayed wire,
		
					wirelist->conn->ept = tempelem;   (tempelem points to the delayed wire) 
		
			while in turn, the delayed wire element's connection list's wire pointer is set to point 
			at the new wire.  
				
					tempelem->conn->wpt = wirelist;
				


			Removal of unconnected physical wires, ordering of indexes, calculation of overall Fanin, 
			Fanout and other variables, and the actual printing of the DEE file are all contained in 
			the main() function below. 

 SIDE EFFECTS:		None
		
 COMMENTS: 		As of now, this translator will only correctly translate gate level Verilog models.  
			Behavioral, transistor, and switch level models will be read in and ignored. 

 HISTORY:		Unmodified

 FUNCTIONS:  		Htelem 	*ht_make(int)	
			Htelem 	*ht_head(Htelem*,Htelem**,Htelem**)
			Htelem 	*ht_tail(Htelem*,Htelem**,Htelem**)
			Htelem 	*ht_unlink(Htelem*,Htelem**,Htelem**)
			int    	 LevelStep(ELEM *elemlist)

*******************************************************************************************************************/

%{
#include <stdio.h>

/* assign integer values to TRUE and FALSE */

#define TRUE  	1
#define FALSE 	0

/* codes defining the position of the wire or element */

#define INTERNAL	0 
#define INP  		1
#define OUTP 		2
#define BIDIR 		3

/* assign integer values to gate types */

#define D_FLIP_FLOP 	44
#define AND_GATE	 0
#define NAND_GATE	 1
#define OR_GATE	 	 2
#define NOR_GATE	 3
#define XOR_GATE	30
#define	XNOR_GATE	31
#define	BUFFER    	 4
#define INVERTER	 5
#define DELAYED_WIRE    99

/* define element (E) type */

#define E_WIRE		 1
#define E_LOGIC_GATE	 5
#define E_BEHAVIORAL 	15

/*******************************************************************************************************************

	The following four macros were taken from Prof. Alva L. Couch's ht list handling library, 
	      Copyright 1992 by Alva L. Couch, Dept. of Computer Science, Tufts University

*******************************************************************************************************************/

#define MALLOC_TYPE(type) (type *) malloc (sizeof(type))

#define HT_HEAD(a,h,t) \
	ht_head((Htelem *)a,(Htelem **)h, (Htelem **)t)
#define HT_TAIL(a,h,t) \
	ht_tail((Htelem *)a,(Htelem **)h, (Htelem **)t)
#define HT_UNLINK(a,h,t) \
	ht_unlink((Htelem *)a, (Htelem **)h, (Htelem **)t)

/******************************************************************************************************************/

int IDnum = 0, wirecount = 0, elemcount = 0, modulecount = 0; 	/*indexes of elem, ID, and module counts*/ 

int i, x;    		 /*indexes*/
int position;	 /*used to store beginning of input and output blocks to set position*/
int cnxn_index;		 /*used to store the link bit of elem connections*/
int delay_index = 1;     /*used to store the index of the delay table*/
int Is_in_list;		 /*Boolean used to denote whether a wire name is in the list of internal wires*/
int PIs = 0, POs = 0, BiDirectionals = 0; 	/*counts of Primary Inputs, Outputs, and Bidirectionals*/

/*******************************************************************************************************************

	Definition of struct used in Prof. Alva L. Couch's ht linked list library

*******************************************************************************************************************/

typedef struct htelem { 
    struct htelem *next; /* next pointer in main list */
    /* anything goes here */ 
} Htelem;

/******************************************************************************************************************/


typedef struct Wire{             	/*used to store internal wire connection information*/

	struct Wire *next;		/*pointer to next wire in list*/

	char *name;			/*name of wire*/

	int fanin;			/*number of gates which fanin to the wire*/

	int fanout;			/*number of gates to which the wire fans out*/ 

	int link_bit; 			/*stores index of output from element (0 if 1 output,*/
					/*0 or 1 if 2 outputs, 0, 1, or 2 if 3 outputs, etc.)*/

	struct ELEM *elem;		/*element (gate or input wire) which the internal wire originates from*/

	struct CONNECTION *couthead;	/*head of output connection list*/

	struct CONNECTION *couttail;	/*tail of output connection list*/

	struct CONNECTION *conn;	/*pointer used to cycle through connection list.  The connection list*/
					/*stores pointers to the elements which accept this wire as an input.*/
					/*Each internal wire stores and maintains its own connection list.   */
} Wire;



typedef struct CONNECTION{             	/*used to store info on one inter-elem connection*/

	struct CONNECTION *next;	/*pointer to next connection in list*/

	struct ELEM *ept;		/*connections which are pointed to by internal wires store a pointer */
					/*to an element.  This element is one of (possibly) many which accept*/
					/*the wire as an input.*/

	struct Wire *wpt;		/* connections which are pointed to by elements store a pointer to an  */
					/*internal wire.  This wire is one of many which are inputs to the elem*/
} CONNECTION;



typedef struct DELAY{   	     	/*delay information*/

	struct DELAY *next;		/*LINKED LIST NOT YET IMPLEMENTED FOR DELAY TABLE*/

	int rise, fall;			/*rise and fall times of each gate.  Gates have an index which*/
					/*points to a certain rise and fall in the delay table.       */
} DELAY;



DELAY delay_table[30];		   	/*delay table*/


	
typedef struct ELEM{	 		/*struct used to store info on wires(inputs and delay wires) + elems*/

	struct ELEM *next;		/*pointer to next element in list*/

	char *name;			/*name of element*/

	int position;			/*O for internal, 1 for input, 2 for output, 3 for bidirectional*/

	int index;			/*our index of the element in the element list*/

	int logictype;			/*type of logic of element (1 if wire, 5 if gate, 15 if behavioral)*/

	int gatetype;			/*if a gate or behavioral element, "gatetype" stores the sub-class */

	int connection;			/*Connection index of the element (UNUSED, as of Aug 3, '98)*/

	int outputs;			/*number of outputs of the element*/

	int inputs;			/*number of inputs which enter the gate*/
	
	int delay;			/*delay index of element -- stores location of the correct*/ 
					/*rise and fall values of the element in the delay table  */

	int sim_lev;			/*simulation level -- the depth of 0-delay elements in the circuit.*/
					/*Input wires are assigned a value of 1 for sim_lev, while gates   */
					/*with a non-zero delay are assigned a value of zero.  All other   */
					/*gates are assigned a value of 1 more than the greatest sim_lev   */
					/*value of all gates which are input into the gate being analyzed. */

	struct CONNECTION *cinhead;	/*head of input connection list*/

	struct CONNECTION *cintail;	/*tail of input connection list*/

	struct CONNECTION *conn;	/*pointer used to cycle through the connection list.  The connection*/
					/*list stores pointers to internal wires which the current element  */
					/*uses as inputs.  Each element maintains its own connection list.  */
} ELEM;



	/***** WIRELIST INSTANTIATION *****/

Wire *wirelist = NULL;		/*Internal wire pointer used for cycling through the wirelist.    */
Wire *whead = NULL;		/*The head and tail of the wirelist are stored in whead and wtail.*/
Wire *wtail = NULL;

	/***** ELEMLIST INSTANTIATION *****/
	
ELEM *elemlist = NULL;		/*Internal element pointer used for cycling through the elemlist. */
ELEM *tempelem = NULL; 			/*Temporary element pointer used for short term storage*/
ELEM *ehead = NULL;		/*The head and tail of the elemlist are stored in ehead and etail.*/	
ELEM *etail = NULL;

Htelem *ht_make(int);	
Htelem *ht_head(Htelem*,Htelem**,Htelem**);
Htelem *ht_tail(Htelem*,Htelem**,Htelem**);
Htelem *ht_unlink(Htelem*,Htelem**,Htelem**);

int LevelStep(ELEM *elemlist);

%}

%union{
	int n;		/*type classes for tokens and non-terminal identifiers (defined below)*/ 
	char *txt;	/*To define a token or non-terminal identifier as a certain class, the*/
	char c;		/*class is denoted in the <> brackets after the %token or %type decl. */
}     

%start source_text	/*begin parsing at the "source_text" non-terminal identifier*/

%token <n>   DECIMAL_NUMBER UNSIGNED_NUMBER EDGE_OR_UNSIGNED ZERO_OR_ONE
%token <txt> BINARY_NUMBER OCTAL_NUMBER HEX_NUMBER REAL_NUMBER
%token <txt> BINARY_OPERATOR UNARY_OPERATOR OPERATOR  
%token <txt> EDGE_DESCRIPTOR EDGE_OR_BINARY_OPERATOR
%token <txt> EX LEVEL_SYMBOL EDGE_SYMBOL
%token <c>   EQUAL OPENPAREN CLOSEPAREN OPENCURLYBRAC CLOSECURLYBRAC BACKSLASH
%token <c>   COMMA OPENSQBRAC CLOSESQBRAC POUND AT COLON SEMICOLON QUESTIONMARK
%token <c>   PLUS MINUS TILDA DOT SINGLE_QUOTE
%token <txt> LESSOREQUAL DOUBLERIGHTARROW EQUIV_EXPR 
%token <txt> RIGHTARROW STARARROWRIGHT
%token <txt> TRIAMPERSAND DOLLAR DES_ASSIGNMENT
%token <txt> KNOWN_BIT ONE_KNOWN_BIT UNKNOWN_BIT
%token <txt> ESCAPED_IDENTIFIER SIMPLE_IDENTIFIER STRING
%token
ALWAYS		/*KEYWORDS WITH NO TYPE*/
AND	
ASSIGN
BEGN
BODY
BUF
BUFIF0 
BUFIF1
CASE
CASEX
CASEZ
CMOS
DEASSIGN
DEFAULT
DEFPARAM
DFF
DISABLE
EDGE
ELSE
END
ENDCASE
ENDFUNCTION
ENDMODULE
ENDPRIMITIVE
ENDSPECIFY
ENDTABLE
ENDTASK
ENDUNIT
EVENT
FOR
FORCE
FOREVER
FORK
FUNCTION
HIGHZ0
HIGHZ1
HOLD
IF
IFNONE
IN
INITL
INOUT
INPUT
INTEGER
INV
JOIN
LARGE
MACROMODULE
MEDIUM
MODULE
NAND
NEGEDGE
NMOS
NOR
NOT
NOTIF0
NOTIF1
OR
OUT
OUTPUT
PARAMETER
PATHPULSE
PERIOD
PMOS
POSEDGE
PRIMITIVE
PULL0
PULL1
PULLDOWN
PULLUP
REAL
REALTIME
RCMOS
RECOVERY
REG
RELEASE
REPEAT
RNMOS
RPMOS
RTRAN
RTRANIF0
RTRANIF1
SCALARED
SETUP
SETUPHOLD
SKEW
SMALL
SPECIFY
SPECPARAM
STRONG0
STRONG1
SUPPLY0
SUPPLY1
TABLE
TASK
TIME
TRAN
TRANIF0
TRANIF1
TRI
TRI0
TRI1
TRIAND
TRIOR
TRIREG
UNIT
VECTORED
WAIT
WAND
WEAK0
WEAK1
WHILE
WIDTH
WIRE
WOR
XNOR
XOR

%type <txt> identifier IDENTIFIER escaped_identifier simple_identifier
%type <n>   delay_value delay_value1 delay_value2 delay_value3 unsigned_number

%%
	/*YACC PARSING SECTION*/

always_construct : ALWAYS statement
		 ; 

binary_number : size BINARY_NUMBER
 	      ;
		
binary_operator : PLUS
		| MINUS
		| OPERATOR
		| EQUIV_EXPR
		| BINARY_OPERATOR
		;

blocking_assignment : reg_lvalue EQUAL delay_or_event_control expression 
		    ;

block_item_declaration  : /*empty*/
			| parameter_declaration 
			| reg_declaration 
			| integer_declaration 
			| real_declaration 
			| time_declaration 
			| realtime_declaration 
			| event_declaration 
			| block_item_declaration parameter_declaration
			| block_item_declaration reg_declaration 
			| block_item_declaration integer_declaration
			| block_item_declaration real_declaration 
			| block_item_declaration time_declaration 
			| block_item_declaration realtime_declaration 
			| block_item_declaration event_declaration
			;
 
case_item : /*empty*/
	  | expression COLON statement_or_null
	  | DEFAULT statement_or_null
	  | DEFAULT COLON statement_or_null 
	  | case_item expression COLON statement_or_null	 
	  | case_item DEFAULT statement_or_null
	  | case_item DEFAULT COLON statement_or_null
	  ;
	
case_statement  : CASE OPENPAREN expression CLOSEPAREN case_item ENDCASE 
		| CASEZ OPENPAREN expression CLOSEPAREN case_item ENDCASE 
  		| CASEX OPENPAREN expression CLOSEPAREN case_item ENDCASE 
		;

charge_strength : /*empty*/
		| OPENPAREN SMALL CLOSEPAREN 
		| OPENPAREN MEDIUM CLOSEPAREN 
		| OPENPAREN LARGE CLOSEPAREN
		;

cmos_switchtype : CMOS 
		| RCMOS  
		; 

cmos_switch_instance : /*empty*/
		     | name_of_gate_instance 
			OPENPAREN output_terminal COMMA input_terminal COMMA 
			ncontrol_terminal COMMA pcontrol_terminal CLOSEPAREN 
		     | cmos_switch_instance COMMA name_of_gate_instance 
			OPENPAREN output_terminal COMMA input_terminal COMMA 
			ncontrol_terminal COMMA pcontrol_terminal CLOSEPAREN 
		     ;

combinational_body : TABLE combinational_entry ENDTABLE 
		   ;

combinational_entry : /*empty*/    
		    | level_input_list COLON output_symbol SEMICOLON
		    | combinational_entry level_input_list COLON output_symbol SEMICOLON
		    ;
	
comment : short_comment 
	| long_comment
	;

comment_text : /*empty*/ /*COMMENTS IGNORED FOR NOW*/
	     ;

concatenation : OPENCURLYBRAC expression CLOSECURLYBRAC 
	      ;

conditional_statement : IF OPENPAREN expression CLOSEPAREN statement_or_null
	  	      | IF OPENPAREN expression CLOSEPAREN statement_or_null 
				ELSE statement_or_null
		      ; 

constant_expression : constant_primary 
		    | unary_operator constant_primary 
		    | constant_expression binary_operator constant_expression 
		    | constant_expression QUESTIONMARK constant_expression 
			COLON constant_expression 
		    | string 
		    ;
	
constant_mintypmax_expression : constant_expression 
			      | constant_expression COLON 
					constant_expression COLON constant_expression 
			      ;
	
constant_primary : number 
		 | identifier 
		 | concatenation 
		 | multiple_concatenation 
		 ;

continuous_assign : ASSIGN drive_strength delay3 list_of_net_assignments SEMICOLON 
		  ;

controlled_timing_check_event :	timing_check_event_control specify_terminal_descriptor 
			      |	timing_check_event_control specify_terminal_descriptor  
				TRIAMPERSAND timing_check_condition  
			      ;

current_state : level_symbol 
	      ;
	
data_source_expression : expression
		       ;	

decimal_number 	: sign UNSIGNED_NUMBER 
		| size DECIMAL_NUMBER 
		;

delay2 	: /*empty*/
		{elemlist->delay = 0;}		/*if no delay specified, set to 0*/
	| POUND delay_value1 
		{delay_index++;
		}
	| POUND OPENPAREN delay_value1 CLOSEPAREN
		{delay_index++;		      
		}
	| POUND OPENPAREN delay_value1 COMMA delay_value3 CLOSEPAREN
	;

delay3 	: /*empty*/
	| POUND delay_value1
		{delay_index++;
		}
	| POUND OPENPAREN delay_value1 CLOSEPAREN
		{delay_index++;
		}
	| POUND OPENPAREN delay_value1 COMMA delay_value3 CLOSEPAREN 
	| POUND OPENPAREN delay_value1 COMMA delay_value2 COMMA delay_value3 CLOSEPAREN
	;
	
delay_control 	: POUND delay_value 
		| POUND OPENPAREN mintypmax_expression CLOSEPAREN
		; 

delay_or_event_control  : /*empty*/
		        | delay_control 
			| event_control 
			| REPEAT OPENPAREN expression CLOSEPAREN event_control
			; 
	
delay_value : unsigned_number 
	    ;
	
delay_value1 : unsigned_number
		{if($1 == 0) elemlist->delay = 0;	/*get delay if != 0*/
		 else {
			x=0;
		 	for(i=1; i<delay_index; i++){ 		
		   	  if(delay_table[i].rise == $1 && delay_table[i].fall == $1){
			    elemlist->delay = i;
			    x++;				/*set to delay index in delay table*/
			  }					/*if delay already exists*/
			}
			if(!x){
			  delay_table[delay_index].rise = $1;	/*create new delay in delay table*/
			  delay_table[delay_index].fall = $1;	/*if delay does not exist*/
			  elemlist->delay = delay_index;
			}
		      }
		} 
	     ;
	
delay_value2 : unsigned_number 
	     ;
	
delay_value3 : unsigned_number 
		{if($1 == 0) elemlist->delay = 0;
		 else if(!x){
		   delay_table[delay_index].fall = $1;	/*if delay does not exist*/
		   delay_index++;
		   }
		}
	     ;
	
description : /*empty*/
	    | module_declaration 
	    | udp_declaration 
	    | des_module_declaration
	    | description module_declaration 
	    | description udp_declaration
	    | description des_module_declaration
	    ;

des_assignment 	: /*empty*/
		| DES_ASSIGNMENT SINGLE_QUOTE ZERO_OR_ONE 
		;

des_delay2 : /*empty*/
		{elemlist->delay = 0;}		/*if no delay specified, set to 0*/
	   | delay_value1
		{delay_index++;
		}
	   | OPENPAREN delay_value1 CLOSEPAREN
		{delay_index++;		      
		}
	   | OPENPAREN delay_value1 COMMA delay_value3 CLOSEPAREN
	   ;

des_delay3 : /*empty*/
	   | POUND delay_value1
		{delay_index++;
		}
	   | OPENPAREN delay_value1 CLOSEPAREN
		{delay_index++;
		}
	   | OPENPAREN delay_value1 COMMA delay_value3 CLOSEPAREN 
	   | OPENPAREN delay_value1 COMMA delay_value2 COMMA delay_value3 CLOSEPAREN
	   ;
	
des_gate_instance : name_of_gate_instance OPENPAREN 
			{elemlist->position = INTERNAL;		/*set gate parameters*/
			 elemlist->logictype = E_LOGIC_GATE;	
			 HT_TAIL(elemlist,&ehead,&etail);
			 elemcount++;
			 IDnum++;}
		    output_terminal_repeat			/*get input and output data*/
		    SEMICOLON
		    input_terminal_repeat
		    CLOSEPAREN
	          | des_gate_instance 
			{elemlist = MALLOC_TYPE(struct ELEM);
			elemlist->gatetype = etail->gatetype;
		  	elemlist->inputs = elemlist->outputs = 0; 
			}
		    COMMA name_of_gate_instance OPENPAREN 
			{elemlist->position = INTERNAL;		/*set gate parameters*/
			 elemlist->logictype = E_LOGIC_GATE;	
			 HT_TAIL(elemlist,&ehead,&etail);
			 elemcount++;
			 IDnum++;}
		    output_terminal_repeat			/*get input and output data*/
		    SEMICOLON
		    input_terminal_repeat
		    CLOSEPAREN
	          ;

des_gate_instantiation 	: n_input_gatetype des_delay2 COLON des_gate_instance SEMICOLON
		   	| n_output_gatetype des_delay2 COLON des_gate_instance SEMICOLON
		   	| dff_gatetype des_delay2 COLON des_gate_instance SEMICOLON
			;

des_module_declaration	: UNIT 
				{modulecount++;}
			  identifier list_of_des_ports des_net_declaration BODY
			  des_module_item ENDUNIT 
			;

des_module_item	: des_gate_instantiation
	        | des_module_item des_gate_instantiation 
		;

des_net_declaration : /*empty*/
		    | net_type vectored_or_scalared range des_delay2 list_of_net_identifiers SEMICOLON 
		    | des_net_declaration net_type vectored_or_scalared range des_delay2 
			list_of_net_identifiers SEMICOLON 
		    ; 

des_port : des_port_expression 
	 | des_port des_port_expression 
	 ;

des_port_expression : IN WIRE 
			{position = INP;}			
		      list_of_port_identifiers des_assignment SEMICOLON
 		    | OUT WIRE 
			{position = OUTP;}
		      list_of_port_identifiers SEMICOLON
		    ;


dff_gatetype : DFF
		{elemlist = MALLOC_TYPE(struct ELEM);
		 elemlist->gatetype = D_FLIP_FLOP;		/*set gatetype*/
		 elemlist->inputs = elemlist->outputs = 0; 
		}
	     ;

dff_gate_instance : name_of_gate_instance OPENPAREN 
			{elemlist->position = INTERNAL;		/*set dff parameters*/
			 elemlist->logictype = E_LOGIC_GATE;	
			 HT_TAIL(elemlist,&ehead,&etail);
			 elemcount++;
			 IDnum++;}
		    input_terminal				/*get first input*/
		    COMMA
		    output_terminal 				/*get output name*/
 		    COMMA 
	 	    dff_input_terminal				/*get second input*/
		    CLOSEPAREN
	          | dff_gate_instance COMMA name_of_gate_instance OPENPAREN 
			{elemlist->position = INTERNAL;		/*set dff parameters*/
			 elemlist->logictype = E_LOGIC_GATE;	
			 HT_TAIL(elemlist,&ehead,&etail);
			 elemcount++;
			 IDnum++;}
		    input_terminal				/*get first input*/
		    COMMA 
		    output_terminal 				/*get output name*/
 		    COMMA 
	 	    dff_input_terminal				/*get second input*/
		    CLOSEPAREN
	          ;

dff_input_terminal: IDENTIFIER
			{Is_in_list=FALSE;
			 for(wirelist=whead; wirelist; wirelist=wirelist->next)
			        if(!strcmp(wirelist->name,$1)){
			           Is_in_list = TRUE;
				   break;
				}
			 if(!Is_in_list){
			   wirelist = MALLOC_TYPE(struct Wire);
			   wirelist->fanin = wirelist->fanout = 0;
			   wirelist->name = $1;
			 }
			 wirelist->fanout++;
			 etail->inputs++;
			 if(!Is_in_list) HT_TAIL(wirelist, &whead, &wtail);
			 wirelist->conn = MALLOC_TYPE(struct CONNECTION);
			 wirelist->conn->ept = etail;
			 elemlist->conn = MALLOC_TYPE(struct CONNECTION);
			 elemlist->conn->wpt = wirelist;
			 HT_HEAD(elemlist->conn, &elemlist->cinhead, &elemlist->cintail);   /*<-link at head instd*/
			 HT_TAIL(wirelist->conn, &wirelist->couthead, &wirelist->couttail); /*of tail to resolve  */
			}								    /*DEE ordering problem*/
 											    /*for DFF inputs      */
		   ; 

disable_statement : DISABLE identifier SEMICOLON
		  ;	

drive_strength 	: /*empty*/
		| OPENPAREN strength0 COMMA strength1 CLOSEPAREN 
		| OPENPAREN strength1 COMMA strength0 CLOSEPAREN 
		| OPENPAREN strength0 COMMA HIGHZ1 CLOSEPAREN 
		| OPENPAREN strength1 COMMA HIGHZ0 CLOSEPAREN 
		| OPENPAREN HIGHZ1 COMMA strength0 CLOSEPAREN 
		| OPENPAREN HIGHZ0 COMMA strength1 CLOSEPAREN 
		;

edge_control_specifier 	: EDGE OPENSQBRAC edge_descriptor CLOSESQBRAC
			| EDGE OPENSQBRAC edge_descriptor COMMA edge_descriptor CLOSESQBRAC 
			;

edge_descriptor : EDGE_OR_UNSIGNED
		| EDGE_DESCRIPTOR
		;
	
edge_identifier : /*empty*/
		| POSEDGE 
		| NEGEDGE
		;

edge_indicator 	: OPENPAREN level_symbol level_symbol CLOSEPAREN 
		| edge_symbol 
		;

edge_input_list : level_symbol edge_indicator level_symbol  
		;

edge_sensitive_path_declaration : parallel_edge_sensitive_path_description 
					EQUAL path_delay_value 
				| full_edge_sensitive_path_description 
					EQUAL path_delay_value
				; 

edge_symbol : EDGE_OR_BINARY_OPERATOR
	    | EDGE_SYMBOL
	    ;

enable_gatetype : BUFIF0 
		| BUFIF1 
		| NOTIF0 
		| NOTIF1
		;

enable_gate_instance : /*empty*/
		     | name_of_gate_instance OPENPAREN output_terminal 
			COMMA input_terminal COMMA enable_terminal CLOSEPAREN 
		     | enable_gate_instance COMMA name_of_gate_instance OPENPAREN 
			output_terminal COMMA input_terminal COMMA enable_terminal 
			CLOSEPAREN  
		     ;

enable_terminal : expression
		;

escaped_identifier : ESCAPED_IDENTIFIER
		   ;

event_control 	: AT identifier 
		| AT OPENPAREN event_expression CLOSEPAREN
  	   	;

event_declaration : EVENT identifier SEMICOLON  
		  ;
	
event_expression : expression 
		 | identifier 
		 | POSEDGE expression 
		 | NEGEDGE expression 
		 | event_expression OR expression 
		 | event_expression OR identifier 
		 | event_expression OR POSEDGE expression 
		 | event_expression OR NEGEDGE expression 
		 ;

event_trigger : RIGHTARROW identifier SEMICOLON
	      ;

expression 	: /*empty*/
		| primary 
		| unary_operator primary 
		| string 
		| expression binary_operator primary 
		| expression binary_operator unary_operator primary 
		| expression binary_operator string 
		| expression QUESTIONMARK expression COLON expression 
		| expression QUESTIONMARK expression COLON expression 
		| expression QUESTIONMARK expression COLON expression 
		;

full_edge_sensitive_path_description : OPENPAREN edge_identifier 
					list_of_path_inputs STARARROWRIGHT 
					list_of_path_outputs 
				 	polarity_operator COLON 
					data_source_expression CLOSEPAREN
				     ; 

full_path_description : OPENPAREN list_of_path_inputs polarity_operator 
			STARARROWRIGHT list_of_path_outputs CLOSEPAREN 
		      ;
		
function_call : identifier OPENPAREN expression CLOSEPAREN 
	      | name_of_system_function 
	      | name_of_system_function OPENPAREN expression CLOSEPAREN 

function_declaration 	: FUNCTION range_or_type identifier SEMICOLON
				function_item_declaration  
				statement 
				ENDFUNCTION 
			;

function_item_declaration : /*empty*/
			  | block_item_declaration 
			  | input_declaration 
			  | function_item_declaration block_item_declaration 
			  | function_item_declaration input_declaration 
			  ;

gate_instantiation : n_input_gatetype drive_strength 
			delay2 n_input_gate_instance SEMICOLON
		   | n_output_gatetype drive_strength delay2  
			n_output_gate_instance SEMICOLON
		   | dff_gatetype drive_strength delay2
			dff_gate_instance SEMICOLON
		   | enable_gatetype drive_strength delay3  
			enable_gate_instance SEMICOLON 
		   | mos_switchtype delay3  
			mos_switch_instance SEMICOLON 
		   | pass_switchtype pass_switch_instance SEMICOLON 
		   | pass_en_switchtype delay3  
			pass_en_switch_instance SEMICOLON 
		   | cmos_switchtype delay3 
			cmos_switch_instance SEMICOLON 
		   | PULLUP pullup_strength 
			pull_gate_instance SEMICOLON 
		   | PULLDOWN pulldown_strength 
			pull_gate_instance SEMICOLON
		   ;

hex_number : size HEX_NUMBER 
	   ;

identifier 	: /*empty*/
			{printf("");}		/*will not run without these empty 'printf's*/		
		| IDENTIFIER
			{printf("");}
		| identifier DOT IDENTIFIER
			{printf("");}
		| IDENTIFIER DOT identifier
			{printf("");}
		;

IDENTIFIER : simple_identifier 
			
	   | escaped_identifier 
			
	   ;

initial_construct : INITL statement 
		  ;

init_val : ZERO_OR_ONE
	 | ONE_KNOWN_BIT
	 | UNKNOWN_BIT
	 ;

inout_declaration : INOUT
			{position = BIDIR;}
			range list_of_port_identifiers SEMICOLON		
		  ; 

inout_terminal : identifier 
	       | identifier OPENSQBRAC constant_expression CLOSESQBRAC
	       ; 

input_declaration : INPUT  
			{position = INP;}			
			range list_of_port_identifiers SEMICOLON	
		  ;

input_identifier : identifier
		 ;

input_port_connection : /*empty*/
		      | expression
		      | input_port_connection COMMA expression
		      ;

input_terminal  : IDENTIFIER
			{Is_in_list=FALSE;
			 for(wirelist=whead; wirelist; wirelist=wirelist->next)
			        if(!strcmp(wirelist->name,$1)){
			           Is_in_list = TRUE;
				   break;
				}
			 if(!Is_in_list){
			   wirelist = MALLOC_TYPE(struct Wire);
			   wirelist->fanin = wirelist->fanout = 0;
			   wirelist->name = $1;
			   wirelist->link_bit = 0;
			 }
			 wirelist->fanout++;
			 elemlist->inputs++;
			 if(!Is_in_list) HT_TAIL(wirelist, &whead, &wtail);
			 wirelist->conn = MALLOC_TYPE(struct CONNECTION);
			 wirelist->conn->ept = elemlist;
			 elemlist->conn = MALLOC_TYPE(struct CONNECTION);
			 elemlist->conn->wpt = wirelist;
			 HT_TAIL(elemlist->conn, &elemlist->cinhead, &elemlist->cintail);
			 HT_TAIL(wirelist->conn, &wirelist->couthead, &wirelist->couttail);
 			}
		;

input_terminal_repeat 	: input_terminal
			| input_terminal_repeat COMMA input_terminal
			;
	
integer_declaration : INTEGER list_of_register_identifiers SEMICOLON
		    ;

level_input_list : level_symbol
		 ;

level_symbol 	: /*empty*/
		| level_symbol ZERO_OR_ONE
		| level_symbol EX
		| level_symbol QUESTIONMARK
		| level_symbol LEVEL_SYMBOL 
		;
	
limit_value : constant_mintypmax_expression 
	    ;

list_of_des_ports :/*empty*/
	          | OPENPAREN des_port CLOSEPAREN
		  ;

list_of_module_connections : /*empty*/
			   | ordered_port_connection
			   | named_port_connection
			   ;

list_of_net_assignments : net_assignment 
			| list_of_net_assignments COMMA net_assignment 
			;

list_of_net_decl_assignments : net_decl_assignment 
			     | list_of_net_decl_assignments COMMA net_decl_assignment
			     ;

list_of_net_identifiers : identifier 
				 {if(elemlist->delay!=0){
				    elemlist->name = $1;
				    elemlist->sim_lev = 0;
				    elemcount++;
				    HT_TAIL(elemlist, &ehead, &etail); 
				    wirelist = MALLOC_TYPE(struct Wire);
			  	    wirelist->name = $1;
				    wirelist->fanin = wirelist->fanout = 0;
				    wirelist->elem = elemlist;
				    elemlist->cinhead = NULL;
				    HT_TAIL(wirelist,&whead,&wtail);
				    }
				 }
			| list_of_net_identifiers COMMA identifier		/*set wire parameters*/
				 {if(elemlist->delay!=0){
				    elemlist->name = $3;
				    elemlist->sim_lev = 0;
				    elemcount++;
				    HT_TAIL(elemlist, &ehead, &etail); 
				    wirelist = MALLOC_TYPE(struct Wire);
			  	    wirelist->name = $3;
				    wirelist->fanin = wirelist->fanout = 0;
				    wirelist->elem = elemlist;
				    elemlist->cinhead = NULL;
				    HT_TAIL(wirelist,&whead,&wtail);
				    }
				 }
			;

list_of_param_assignments : param_assignment 
			  | list_of_param_assignments COMMA param_assignment 
			  ;

list_of_path_delay_expressions  : path_delay_expression 
			        | path_delay_expression COMMA path_delay_expression 
				| path_delay_expression COMMA path_delay_expression 
					COMMA path_delay_expression 
				| path_delay_expression COMMA path_delay_expression 
					COMMA path_delay_expression COMMA 
					path_delay_expression COMMA path_delay_expression 
					COMMA path_delay_expression 
				| path_delay_expression COMMA path_delay_expression 
					COMMA path_delay_expression COMMA
					path_delay_expression COMMA path_delay_expression 
					COMMA path_delay_expression COMMA 
					path_delay_expression COMMA path_delay_expression 
					COMMA path_delay_expression COMMA 
					path_delay_expression COMMA path_delay_expression 
					COMMA path_delay_expression
				; 

list_of_path_inputs : specify_input_terminal_descriptor 
		    ;

list_of_path_outputs : specify_output_terminal_descriptor 
		     ;

list_of_ports : /*empty*/
	      | OPENPAREN port CLOSEPAREN
	      ;

list_of_port_identifiers : identifier
				{elemlist = MALLOC_TYPE(struct ELEM);
				 elemlist->index = IDnum;	/*set input and output ports*/
				 elemlist->logictype = E_WIRE;	/*with wire parameters*/
				 elemlist->name = $1;
				  elemlist->inputs = 1;
				  elemlist->outputs = 1;
				 elemlist->position = position;
				 HT_TAIL(elemlist,&ehead,&etail); 
				 wirelist = MALLOC_TYPE(struct Wire);
			  	  wirelist->name = $1;
				 wirelist->fanin = wirelist->fanout = 0;
				 if(wirelist->elem==NULL) wirelist->elem = elemlist;
				   elemlist->cinhead = NULL;
				 HT_TAIL(wirelist,&whead,&wtail);
				 if(position == INP) {
				   PIs++;
			 	   elemlist->sim_lev = 1;
				   elemlist->cinhead = NULL;
				 }
				 else if(position == BIDIR) BiDirectionals++;
				 else if(position == OUTP) POs++;	
				 elemcount++;
				 IDnum++;}
			 | list_of_port_identifiers COMMA identifier 	/*set an input or output port*/
				{elemlist = MALLOC_TYPE(struct ELEM);
				 elemlist->index = IDnum;	/*set input and output ports*/
				 elemlist->logictype = E_WIRE;	/*with wire parameters*/
				 elemlist->name = $3;
				  elemlist->inputs = 1;
				  elemlist->outputs = 1;
				 elemlist->position = position;
				 HT_TAIL(elemlist,&ehead,&etail); 
				 wirelist = MALLOC_TYPE(struct Wire);
			  	  wirelist->name = $3;
				 wirelist->fanin = wirelist->fanout = 0;
				 if(wirelist->elem==NULL) wirelist->elem = elemlist;
				   elemlist->cinhead = NULL;
				 HT_TAIL(wirelist,&whead,&wtail);
				 if(position == INP) {
				   PIs++;
			 	   elemlist->sim_lev = 1;
				   elemlist->cinhead = NULL;
				 }
				 else if(position == BIDIR) BiDirectionals++;
				 else if(position == OUTP) POs++;		
				 elemcount++;
				 IDnum++;}
			 ;

list_of_real_identifiers : identifier
			 ;

list_of_register_identifiers : register_name
			     ;

list_of_specparam_assignments : specparam_assignment 
			      ;

long_comment : /*COMMENTS IGNORED FOR NOW*/
	     ;

loop_statement 	: FOREVER statement 
		| REPEAT OPENPAREN expression CLOSEPAREN statement 
		| WHILE OPENPAREN expression CLOSEPAREN statement 
		| FOR OPENPAREN reg_assignment SEMICOLON expression SEMICOLON 
			reg_assignment CLOSEPAREN statement 
		;

mintypmax_expression : expression
		     | expression COLON expression COLON expression 
		     ;

module_declaration : module_keyword identifier list_of_ports SEMICOLON 
			module_item ENDMODULE 
		   ;

module_instance : /*empty*/
		| name_of_instance OPENPAREN list_of_module_connections CLOSEPAREN 
		| module_instance COMMA name_of_instance OPENPAREN 
			list_of_module_connections CLOSEPAREN
		;

module_instantiation : identifier parameter_value_assignment 
			module_instance SEMICOLON

module_item : /*empty*/
	    | module_item_declaration 
	    | parameter_override 
	    | continuous_assign 
	    | gate_instantiation 
	    | udp_instantiation 
	    | module_instantiation 
            | specify_block 
            | initial_construct 
            | always_construct 
	    | module_item module_item_declaration 
	    | module_item parameter_override 
	    | module_item continuous_assign 
	    | module_item gate_instantiation 
	    | module_item udp_instantiation 
	    | module_item module_instantiation 
	    | module_item specify_block 
	    | module_item initial_construct 
	    | module_item always_construct 
	    ;

module_item_declaration : parameter_declaration 
			| input_declaration 
			| output_declaration 
			| inout_declaration 
			| net_declaration 
			| reg_declaration 
			| integer_declaration 
			| real_declaration 
			| time_declaration 
			| realtime_declaration 
			| event_declaration 
			| task_declaration 
			| function_declaration 
			;

module_keyword : MODULE 
			{modulecount++;}  		/*increment the module count*/
	       | MACROMODULE
	       ;

mos_switchtype 	: NMOS
	       	| PMOS
		| RNMOS
		| RPMOS
		;

mos_switch_instance : /*empty*/
		    | name_of_gate_instance OPENPAREN output_terminal COMMA 
			input_terminal COMMA enable_terminal CLOSEPAREN
		    | mos_switch_instance COMMA name_of_gate_instance OPENPAREN 
			output_terminal COMMA input_terminal COMMA enable_terminal 
			CLOSEPAREN
		    ; 

multiple_concatenation : OPENCURLYBRAC expression OPENCURLYBRAC expression 
				CLOSECURLYBRAC CLOSECURLYBRAC
		       ;	 

named_port_connection : /*empty*/
		      | identifier OPENPAREN expression CLOSEPAREN 
		      | named_port_connection COMMA identifier OPENPAREN expression 
			CLOSEPAREN 
		      ;

name_of_gate_instance : /*empty*/	
		      | identifier range
				{
				elemlist->name = $1; 		/*get element name and index*/
				elemlist->index = IDnum;}
		      ;	 

name_of_instance : module_instance identifier range
		 ;

name_of_system_function : DOLLAR identifier 
			;

name_of_udp_instance : /*empty*/
	  	     | identifier range
		     ;

ncontrol_terminal : expression
		  ;

net_assignment 	: /*empty*/
		| net_lvalue EQUAL expression 
		| net_assignment COMMA net_lvalue EQUAL expression 
		;

net_declaration : /*empty*/
		| net_type vectored_or_scalared range delay3 list_of_net_identifiers SEMICOLON 
		; 

net_decl_assignment : identifier EQUAL expression 
		    ;

net_lvalue : identifier 
	   | identifier OPENSQBRAC expression CLOSESQBRAC 
	   | identifier OPENSQBRAC constant_expression COLON constant_expression 
		CLOSESQBRAC
	   | concatenation 
	   ;

net_type : WIRE
		{elemlist = MALLOC_TYPE(struct ELEM);
		 elemlist->logictype = E_LOGIC_GATE;
		 elemlist->gatetype = DELAYED_WIRE;
		 elemlist->inputs = elemlist->outputs = 1; 	
		 elemlist->delay = 0;
		}
	 | TRI 
	 | TRI1 
 	 | SUPPLY0 
	 | WAND
	 | TRIAND 
	 | TRI0 
	 | SUPPLY1 
	 | WOR
	 | TRIOR 
	 ;

next_state : output_symbol 
	   | MINUS
	   ;

non_blocking_assignment : reg_lvalue LESSOREQUAL delay_or_event_control expression 
			;

notify_register : identifier 
		;

number 	: decimal_number 
	| octal_number 
	| binary_number 	
	| hex_number 
	| real_number 
	| unsigned_number
	| ZERO_OR_ONE
	;

n_input_gatetype : AND
			{elemlist = MALLOC_TYPE(struct ELEM);
			elemlist->gatetype = AND_GATE;	
		 	elemlist->inputs = elemlist->outputs = 0; /*For each gate type, assign a value*/
			}
		 | NAND
			{elemlist = MALLOC_TYPE(struct ELEM);
			elemlist->gatetype = NAND_GATE;
		  	elemlist->inputs = elemlist->outputs = 0; 
			}
		 | OR 
			{elemlist = MALLOC_TYPE(struct ELEM);
			elemlist->gatetype = OR_GATE;
		  	elemlist->inputs = elemlist->outputs = 0; 
			}
		 | NOR 
			{elemlist = MALLOC_TYPE(struct ELEM);
			elemlist->gatetype = NOR_GATE;
		  	elemlist->inputs = elemlist->outputs = 0; 
			}
		 | XOR 
			{elemlist = MALLOC_TYPE(struct ELEM);
			elemlist->gatetype = XOR_GATE;
		  	elemlist->inputs = elemlist->outputs = 0; 
			}
		 | XNOR
			{elemlist = MALLOC_TYPE(struct ELEM);
			elemlist->gatetype = XNOR_GATE;
		  	elemlist->inputs = elemlist->outputs = 0; 
			}
		 ; 

n_input_gate_instance : name_of_gate_instance OPENPAREN 
				{elemlist->position = INTERNAL;		/*set gate parameters*/
				 elemlist->logictype = E_LOGIC_GATE;
				 HT_TAIL(elemlist,&ehead,&etail); 
				elemcount++;
				IDnum++;}
			   output_terminal 				/*get name of output*/
			   COMMA 
		 	   input_terminal_repeat 			/*get names of inputs in list*/
			   CLOSEPAREN
		      | n_input_gate_instance 
				{elemlist = MALLOC_TYPE(struct ELEM);
				 elemlist->gatetype = etail->gatetype;
		  		 elemlist->inputs = elemlist->outputs = 0; 
				}
			COMMA name_of_gate_instance OPENPAREN 
				{elemlist->position = INTERNAL;		/*set gate parameters*/
				 elemlist->logictype = E_LOGIC_GATE;
				 HT_TAIL(elemlist,&ehead,&etail); 
				elemcount++;
				IDnum++;}
			   output_terminal 				/*get name of output*/
			   COMMA 
		 	   input_terminal_repeat 			/*get names of inputs in list*/
			   CLOSEPAREN
		      ;

n_output_gatetype : 
		  | BUF 
			{elemlist = MALLOC_TYPE(struct ELEM);
			elemlist->gatetype = BUFFER;
		  	elemlist->inputs = elemlist->outputs = 0; 	/*for each gate type, assign*/
			}						/*a value and increment that*/
		  | NOT							/*gate's gate count.*/
			{elemlist = MALLOC_TYPE(struct ELEM);
			elemlist->gatetype = INVERTER;
		  	elemlist->inputs = elemlist->outputs = 0; 
			}
		  | INV						
			{elemlist = MALLOC_TYPE(struct ELEM);
			elemlist->gatetype = INVERTER;
		  	elemlist->inputs = elemlist->outputs = 0; 
			}
		  ; 

n_output_gate_instance : name_of_gate_instance OPENPAREN 
				{elemlist->position = INTERNAL;		/*set gate parameters*/
				 elemlist->logictype = E_LOGIC_GATE;
				 HT_TAIL(elemlist,&ehead,&etail); 
				elemcount++;
				IDnum++;}
			   output_terminal
			   COMMA input_terminal	 
			   CLOSEPAREN
		       | n_output_gate_instance 						
				{elemlist = MALLOC_TYPE(struct ELEM);
				 elemlist->gatetype = etail->gatetype;
		  		 elemlist->inputs = elemlist->outputs = 0; 
				}
			 COMMA name_of_gate_instance OPENPAREN 
		 	   	{elemlist->position = INTERNAL;		/*set gate parameters*/
				 elemlist->logictype = E_LOGIC_GATE;
				 HT_TAIL(elemlist,&ehead,&etail); 
				elemcount++;
				IDnum++;}
			 output_terminal
			 COMMA input_terminal	 
			 CLOSEPAREN
		       ;	 

octal_number : size OCTAL_NUMBER
	     ;

ordered_port_connection : /*empty*/
			| ordered_port_connection COMMA ordered_port_connection
			| expression	
			; 

output_declaration : OUTPUT 
			{position = OUTP;}
			range list_of_port_identifiers SEMICOLON
		   ; 

output_identifier : identifier
		  ;	

output_port_connection : /*empty*/
 		       | expression
		       ;

output_symbol : ZERO_OR_ONE
	      | EX
	      ;

output_terminal : IDENTIFIER 
			{Is_in_list=FALSE;
			 for(wirelist=whead; wirelist; wirelist=wirelist->next)
			        if(!strcmp(wirelist->name,$1)){
			           Is_in_list = TRUE;
				   break;
				}
			 if(!Is_in_list){
			   wirelist = MALLOC_TYPE(struct Wire);
			   wirelist->fanin = wirelist->fanout = 0;
			   wirelist->name = $1;
			 }
			 wirelist->fanin++;
			 wirelist->link_bit = elemlist->outputs;
			 elemlist->outputs++;
			 if(wirelist->elem != NULL){      /*check for inserted delayed wire*/
			   if(wirelist->elem->gatetype == DELAYED_WIRE){
				tempelem = wirelist->elem;		 /*in the case of a delayed wire,*/
				wirelist = MALLOC_TYPE(struct Wire);	 /*create a new internal wire for*/
			   	wirelist->fanin = wirelist->fanout = 1;	 /*the inter-elem connection (this*/
			   	wirelist->name = NULL;			 /*wire is not declared in the source*/ 
				wirelist->conn = MALLOC_TYPE(struct CONNECTION);   /*file, so it must be  */
				wirelist->conn->ept = tempelem;			   /*created here)*/	
			 	tempelem->conn = MALLOC_TYPE(struct CONNECTION);
			 	tempelem->conn->wpt = wirelist;
				HT_TAIL(wirelist, &whead, &wtail);
			 	HT_TAIL(tempelem->conn, &tempelem->cinhead, &tempelem->cintail);
			 	HT_TAIL(wirelist->conn, &wirelist->couthead, &wirelist->couttail);
			   }
			 }
			 wirelist->elem = elemlist;
			 if(!Is_in_list) HT_TAIL(wirelist, &whead, &wtail);
			}
		| IDENTIFIER OPENSQBRAC constant_expression CLOSESQBRAC 
			{Is_in_list=FALSE;
			 for(wirelist=whead; wirelist; wirelist=wirelist->next)
			        if(!strcmp(wirelist->name,$1)){
			           Is_in_list = TRUE;
				   break;
				}
			 if(!Is_in_list){
			   wirelist = MALLOC_TYPE(struct Wire);
			   wirelist->fanin = wirelist->fanout = 0;
			   wirelist->name = $1;
			 }
			 wirelist->fanin++;
			 wirelist->link_bit = elemlist->outputs;
			 elemlist->outputs++;
			 if(wirelist->elem != NULL){      /*check for inserted delayed wire*/
			   if(wirelist->elem->gatetype == DELAYED_WIRE){
				tempelem = wirelist->elem;		 /*in the case of a delayed wire,*/
				wirelist = MALLOC_TYPE(struct Wire);	 /*create a new internal wire for*/
			   	wirelist->fanin = wirelist->fanout = 1;	 /*the inter-elem connection (this*/
			   	wirelist->name = NULL;			 /*wire is not declared in the source*/ 
				wirelist->conn = MALLOC_TYPE(struct CONNECTION);   /*file, so it must be  */
				wirelist->conn->ept = tempelem;			   /*created here)*/	
			 	tempelem->conn = MALLOC_TYPE(struct CONNECTION);
			 	tempelem->conn->wpt = wirelist;
				HT_TAIL(wirelist, &whead, &wtail);
			 	HT_TAIL(tempelem->conn, &tempelem->cinhead, &tempelem->cintail);
			 	HT_TAIL(wirelist->conn, &wirelist->couthead, &wirelist->couttail);
			   }
			 }
			 wirelist->elem = elemlist;
			 if(!Is_in_list) HT_TAIL(wirelist, &whead, &wtail);
			}
		;

output_terminal_repeat 	: output_terminal
			| output_terminal_repeat COMMA output_terminal
			;
 
parallel_edge_sensitive_path_description : OPENPAREN edge_identifier 
					   specify_input_terminal_descriptor DOUBLERIGHTARROW 
					   specify_output_terminal_descriptor 
					   polarity_operator COLON data_source_expression 
					   CLOSEPAREN  
					 ;

parallel_path_description : OPENPAREN specify_input_terminal_descriptor 
				polarity_operator DOUBLERIGHTARROW 
				specify_output_terminal_descriptor CLOSEPAREN
			  ; 

parameter_declaration : PARAMETER list_of_param_assignments SEMICOLON
		      ; 

parameter_override : DEFPARAM list_of_param_assignments SEMICOLON 
		   ;

parameter_value_assignment : /*empty*/
			   | POUND OPENPAREN expression CLOSEPAREN
			   ;

param_assignment : identifier EQUAL constant_expression 
		 | param_assignment COMMA identifier EQUAL constant_expression 
		 ;

par_block : FORK statement JOIN
	  | FORK COLON identifier block_item_declaration statement JOIN 
	  ;

pass_en_switch_instance : /*empty*/
			| name_of_gate_instance OPENPAREN inout_terminal COMMA 
				inout_terminal COMMA enable_terminal CLOSEPAREN
			| pass_en_switch_instance COMMA name_of_gate_instance OPENPAREN 
				inout_terminal COMMA inout_terminal COMMA enable_terminal 
				CLOSEPAREN
			; 

pass_en_switchtype : TRANIF0 
		   | TRANIF1 
		   | RTRANIF1 
		   | RTRANIF0 
		   ;

pass_switchtype : TRAN 
		| RTRAN
		;

pass_switch_instance : name_of_gate_instance OPENPAREN inout_terminal 
			COMMA inout_terminal CLOSEPAREN 
		     ;

path_declaration : simple_path_declaration SEMICOLON
		 | edge_sensitive_path_declaration SEMICOLON 
		 | state_dependent_path_declaration SEMICOLON
		 ; 

path_delay_expression : constant_mintypmax_expression 
		      ;

path_delay_value : list_of_path_delay_expressions 
		 | OPENPAREN list_of_path_delay_expressions CLOSEPAREN 
		 ;

pcontrol_terminal : expression 
		  ;

polarity_operator : /*empty*/
		  | PLUS
		  | MINUS
		  ;

port 	: port_expression 
	| identifier OPENPAREN port_expression CLOSEPAREN 
	| port COMMA port_expression 
	| port COMMA identifier OPENPAREN port_expression CLOSEPAREN 
	;

port_expression : /*empty*/
		| port_reference 
		| OPENCURLYBRAC port_reference CLOSECURLYBRAC
		;

port_reference 	: /*empty*/
		| identifier
		| identifier OPENSQBRAC constant_expression CLOSESQBRAC 
		| identifier OPENSQBRAC constant_expression COLON 
			constant_expression CLOSESQBRAC 
		| port_reference COMMA identifier
		| port_reference COMMA identifier OPENSQBRAC constant_expression 
			CLOSESQBRAC 
		| port_reference COMMA identifier OPENSQBRAC constant_expression COLON 
			constant_expression CLOSESQBRAC 
		;

primary : number 
	| identifier 
	| identifier OPENSQBRAC expression CLOSESQBRAC
	| identifier OPENSQBRAC constant_expression COLON constant_expression CLOSESQBRAC 
	| concatenation 
	| multiple_concatenation 
	| function_call 
	| OPENPAREN mintypmax_expression CLOSEPAREN 
	;

procedural_continuous_assignment : ASSIGN reg_assignment SEMICOLON
				 | DEASSIGN reg_lvalue SEMICOLON 
				 | FORCE reg_assignment SEMICOLON 
				 | FORCE net_assignment SEMICOLON 
				 | RELEASE reg_lvalue SEMICOLON 
				 | RELEASE net_lvalue SEMICOLON 
				 ;

procedural_timing_control_statement : delay_or_event_control statement_or_null
				    ; 

pulldown_strength : /*empty*/
		  | OPENPAREN strength0 COMMA strength1 CLOSEPAREN
		  | OPENPAREN strength1 COMMA strength0 CLOSEPAREN 
		  | OPENPAREN strength0 CLOSEPAREN 
		  ;

pullup_strength : /*empty*/
		| OPENPAREN strength0 COMMA strength1 CLOSEPAREN
		| OPENPAREN strength1 COMMA strength0 CLOSEPAREN 
		| OPENPAREN strength1 CLOSEPAREN 
		;

pull_gate_instance : /*empty*/
		   | name_of_gate_instance OPENPAREN output_terminal CLOSEPAREN
		   | pull_gate_instance COMMA name_of_gate_instance OPENPAREN 
			output_terminal CLOSEPAREN
		   ;

pulse_control_specparam : PATHPULSE DOLLAR EQUAL OPENPAREN limit_value 
				CLOSEPAREN SEMICOLON 
			| PATHPULSE DOLLAR EQUAL OPENPAREN limit_value COMMA limit_value 
				CLOSEPAREN SEMICOLON 
			| PATHPULSE DOLLAR specify_input_terminal_descriptor DOLLAR 
				specify_output_terminal_descriptor 
				EQUAL limit_value CLOSEPAREN SEMICOLON 
			| PATHPULSE DOLLAR specify_input_terminal_descriptor DOLLAR 
				specify_output_terminal_descriptor 
				EQUAL limit_value COMMA limit_value CLOSEPAREN SEMICOLON
			;

range 	: /*empty*/
	| OPENSQBRAC constant_expression COLON constant_expression CLOSESQBRAC 
	;

range_or_type 	: /*empty*/
		| range 
		| INTEGER 
		| REAL 
		| REALTIME 
		| TIME
		;

realtime_declaration : REALTIME list_of_real_identifiers SEMICOLON
		     ;

real_declaration : REAL list_of_real_identifiers SEMICOLON
		 ;

real_number : sign REAL_NUMBER
	    ;

register_name 	: /*empty*/
		| identifier 
		| identifier OPENSQBRAC constant_expression COLON
			constant_expression CLOSESQBRAC 
	 	| register_name COMMA identifier 
	 	| register_name COMMA identifier OPENSQBRAC constant_expression COLON
			constant_expression CLOSESQBRAC 
		;

reg_assignment : reg_lvalue EQUAL expression 
	       ;

reg_declaration : REG range list_of_register_identifiers SEMICOLON
		;

reg_lvalue : identifier 
	   | identifier OPENSQBRAC expression OPENSQBRAC 
	   | identifier OPENSQBRAC constant_expression COLON constant_expression CLOSESQBRAC 
	   | concatenation 
	   ;

scalar_constant : ZERO_OR_ONE
		| ONE_KNOWN_BIT
		| KNOWN_BIT
		;

scalar_timing_check_condition : expression 
			      | TILDA expression 
			      | expression EQUIV_EXPR scalar_constant 
			      ;

sequential_body : udp_initial_statement TABLE sequential_entry ENDTABLE 

sequential_entry : /*empty*/
		 | seq_input_list COLON current_state COLON next_state SEMICOLON
		 | sequential_entry seq_input_list COLON current_state COLON next_state 
			SEMICOLON
		 ; 

seq_block : BEGN COLON identifier block_item_declaration statement END 
	  | BEGN statement END 
	  ;

seq_input_list 	: level_input_list 
		| edge_input_list 
		;

short_comment : /*COMMENTS IGNORED FOR NOW*/
	      ;

sign : /*empty*/
     | PLUS
     | MINUS 
     ;

simple_identifier : SIMPLE_IDENTIFIER
		  ;

simple_path_declaration : parallel_path_description EQUAL path_delay_value 
			| full_path_description EQUAL path_delay_value 
			;

size : /*empty*/
     | unsigned_number 
     ;

source_text : description
	    ;

specify_block : SPECIFY specify_item ENDSPECIFY 
	      ;

specify_input_terminal_descriptor : input_identifier 
				  | input_identifier OPENSQBRAC constant_expression 
					CLOSESQBRAC 
				  | input_identifier OPENSQBRAC constant_expression 
					COLON constant_expression CLOSESQBRAC
				  | specify_input_terminal_descriptor COMMA 
					input_identifier 
				  | specify_input_terminal_descriptor COMMA 
					input_identifier OPENSQBRAC constant_expression 
					CLOSESQBRAC 
				  | specify_input_terminal_descriptor COMMA 
					input_identifier OPENSQBRAC constant_expression 
					COLON constant_expression CLOSESQBRAC
				  ; 

specify_item : /*empty*/
	     | specparam_declaration 
	     | path_declaration 
	     | system_timing_check 
	     ;

specify_output_terminal_descriptor : output_identifier 
				   | output_identifier OPENSQBRAC constant_expression 
					CLOSESQBRAC 
				   | output_identifier OPENSQBRAC constant_expression 
					COLON constant_expression CLOSESQBRAC
				   | specify_output_terminal_descriptor COMMA 
					output_identifier 
				   | specify_output_terminal_descriptor COMMA 
					output_identifier OPENSQBRAC constant_expression 
					CLOSESQBRAC 
				   | specify_output_terminal_descriptor COMMA 		   
					output_identifier OPENSQBRAC constant_expression 
					COLON constant_expression CLOSESQBRAC
				   ; 

specify_terminal_descriptor : specify_input_terminal_descriptor 
			    | specify_output_terminal_descriptor 
			    ;

specparam_assignment : /*empty*/
		     | identifier EQUAL constant_expression 
		     | pulse_control_specparam 
		     | specparam_assignment COMMA identifier EQUAL constant_expression 
		     | specparam_assignment COMMA pulse_control_specparam 
		     ;

specparam_declaration : SPECPARAM list_of_specparam_assignments SEMICOLON 

statement : /*empty*/
	  | blocking_assignment SEMICOLON 
	  | non_blocking_assignment SEMICOLON 
	  | procedural_continuous_assignment SEMICOLON 
	  | procedural_timing_control_statement 
	  | conditional_statement 
	  | case_statement 
	  | loop_statement 
	  | wait_statement 
	  | disable_statement 
	  | event_trigger 
	  | seq_block 
	  | par_block 
	  | task_enable 
	  | system_task_enable 
	  | statement blocking_assignment SEMICOLON 
	  | statement non_blocking_assignment SEMICOLON 
	  | statement procedural_continuous_assignment SEMICOLON 
	  | statement procedural_timing_control_statement 
	  | statement conditional_statement 
	  | statement case_statement 
	  | statement loop_statement 
	  | statement wait_statement 
	  | statement disable_statement 
	  | statement event_trigger 
	  | statement seq_block 
	  | statement par_block 
	  | statement task_enable 
	  | statement system_task_enable 
	  ;

statement_or_null : statement 
		  | SEMICOLON
		  ; 

state_dependent_path_declaration : IF OPENPAREN expression CLOSEPAREN
				 	simple_path_declaration 
				 | IF OPENPAREN expression CLOSEPAREN 
					edge_sensitive_path_declaration 
				 | IFNONE simple_path_declaration
				 ;

strength0 : SUPPLY0 
	  | STRONG0 
	  | PULL0 
	  | WEAK0
	  ;

strength1 : SUPPLY1 
	  | STRONG1 
	  | PULL1 
	  | WEAK1 
	  ;
		
string 	: STRING
	;

system_task_enable : system_task_name SEMICOLON
		   | system_task_name OPENPAREN expression CLOSEPAREN SEMICOLON
		   ;

system_task_name : DOLLAR identifier 
		 ;

system_timing_check : DOLLAR SETUP OPENPAREN timing_check_event COMMA timing_check_event 
			COMMA timing_check_limit CLOSEPAREN SEMICOLON 
		    | DOLLAR SETUP OPENPAREN timing_check_event COMMA timing_check_event 
			COMMA timing_check_limit COMMA notify_register CLOSEPAREN SEMICOLON
		    | DOLLAR HOLD OPENPAREN timing_check_event COMMA timing_check_event 
			COMMA timing_check_limit CLOSEPAREN SEMICOLON
		    | DOLLAR HOLD OPENPAREN timing_check_event COMMA timing_check_event 
			COMMA timing_check_limit COMMA notify_register CLOSEPAREN SEMICOLON 
		    | DOLLAR PERIOD OPENPAREN controlled_timing_check_event COMMA 
			timing_check_limit CLOSEPAREN SEMICOLON 
		    | DOLLAR PERIOD OPENPAREN controlled_timing_check_event COMMA 
			timing_check_limit COMMA notify_register CLOSEPAREN SEMICOLON 
		    | DOLLAR WIDTH OPENPAREN controlled_timing_check_event COMMA 
			timing_check_limit COMMA constant_expression CLOSEPAREN SEMICOLON 
		    | DOLLAR WIDTH OPENPAREN controlled_timing_check_event COMMA 
			timing_check_limit COMMA constant_expression COMMA notify_register 
			CLOSEPAREN SEMICOLON 
		    | DOLLAR SKEW OPENPAREN timing_check_event COMMA timing_check_event 
			COMMA timing_check_limit CLOSEPAREN SEMICOLON 
		    | DOLLAR SKEW OPENPAREN timing_check_event COMMA timing_check_event 
			COMMA timing_check_limit COMMA notify_register CLOSEPAREN SEMICOLON 
		    | DOLLAR RECOVERY OPENPAREN controlled_timing_check_event COMMA 
			timing_check_event COMMA timing_check_limit CLOSEPAREN SEMICOLON 
		    | DOLLAR RECOVERY OPENPAREN controlled_timing_check_event COMMA 
			timing_check_event COMMA timing_check_limit COMMA notify_register 
			CLOSEPAREN SEMICOLON
		    | DOLLAR SETUPHOLD OPENPAREN timing_check_event COMMA timing_check_event
			COMMA timing_check_limit COMMA timing_check_limit CLOSEPAREN 
			SEMICOLON 
		    | DOLLAR SETUPHOLD OPENPAREN timing_check_event COMMA timing_check_event
			COMMA timing_check_limit COMMA timing_check_limit COMMA 
			notify_register CLOSEPAREN SEMICOLON
		    ;

task_argument_declaration : /*empty*/
			  | block_item_declaration 
			  | output_declaration 
			  | inout_declaration 
			  | task_argument_declaration block_item_declaration 
			  | task_argument_declaration output_declaration 
			  | task_argument_declaration inout_declaration 
			  ;

task_declaration : TASK identifier SEMICOLON task_argument_declaration 
			statement_or_null ENDTASK 
		 ;

task_enable : identifier OPENPAREN expression CLOSEPAREN SEMICOLON 
	    ;

time_declaration : TIME list_of_register_identifiers SEMICOLON

timing_check_condition 	: scalar_timing_check_condition 
			| OPENPAREN scalar_timing_check_condition CLOSEPAREN 
			;

timing_check_event : timing_check_event_control specify_terminal_descriptor 
		   | timing_check_event_control specify_terminal_descriptor
			TRIAMPERSAND timing_check_condition 
		   ; 

timing_check_event_control : /*empty*/
			   | POSEDGE 
			   | NEGEDGE
			   | edge_control_specifier 
			   ;

timing_check_limit : expression 
		   ;

udp_body : combinational_body 
	 | sequential_body
	 ;

udp_declaration : PRIMITIVE identifier OPENPAREN udp_port_list CLOSEPAREN SEMICOLON 
			udp_port_declaration udp_body 
			ENDPRIMITIVE 
		;

udp_initial_statement : /*empty*/
		      | INITL identifier EQUAL init_val SEMICOLON 
		      ;

udp_instance : /*empty*/
  	     | name_of_udp_instance OPENPAREN output_port_connection COMMA 
		input_port_connection CLOSEPAREN
	     | udp_instance COMMA name_of_udp_instance OPENPAREN output_port_connection 
		COMMA input_port_connection CLOSEPAREN
	     ;

udp_instantiation : identifier drive_strength delay2 udp_instance SEMICOLON
		  ; 

udp_port_declaration : /*empty*/
		     | output_declaration 
		     | input_declaration 
		     | reg_declaration 
		     | udp_port_declaration output_declaration 
		     | udp_port_declaration input_declaration 
		     | udp_port_declaration reg_declaration 
		     ;

udp_port_list : identifier COMMA identifier 
	      ;

unary_operator 	: PLUS
		| MINUS
		| OPERATOR
		| TILDA
		| UNARY_OPERATOR
		;

unsigned_number : EDGE_OR_UNSIGNED
		| ZERO_OR_ONE
		| UNSIGNED_NUMBER
		; 

vectored_or_scalared : /*empty*/
		     | VECTORED
		     | SCALARED
		     ;

wait_statement 	: WAIT OPENPAREN expression CLOSEPAREN statement_or_null 
		;

	/*END OF YACC PARSING SECTION*/
%%

#define NET_EXTENSION ".dee"
#define DES_EXTENSION ".des"
#define VERILOG_EXTENSION ".v" 

extern FILE *yyin, *yyout;
extern int yydebug;

/*******************************************************************************************************************

 FILENAME:		VERILOG.YACC

 FUNCTION NAME: 	main(int argc, char** argv)

 CREATION DATE:		August 3, 1998

 AUTHOR:		Ben Foster

 PARAMETERS:	Input:  Name of Verilog file to be translated (minus the ".v" extension)
			Name of DEE file to be created and written to (minus the ".dee" extension)

 PURPOSE:		Accepts the names of the Verilog file to be read and the DEE file to be written to.  
			The main function opens the Verilog file as read only and the DEE file as write to, 
			calls the yyparse() function to begin parsing the Verilog file, and closes the Verilog 
			file after the parsing has finished.  "Main()" then analyzes the data structures 
			created by the yacc routine, modifying them as necessary and finally printing out 
			the DEE file using these data structures.
 
 SIDE EFFECTS:		The data structures "wirelist" and "elemlist" are modified as needed in order to
			clean up the DEE printing.  Loose wires are removed, and some values, such as
			"sim_lev", are modified in the "main()" function.

 COMMENTS: 		

 HISTORY:		Unmodified

*******************************************************************************************************************/

main(argc, argv)
int argc;
char** argv;
{
  char* infilename;		/*strings for input and output files. */
  char* outfilename;
  
  int mod_tb_addr = 0, sym_tb_addr = 0;     	/*pointers to module address and SMT address*/

  int Inputs=0, Outputs=0, Fanin=0, Fanout=0;  	/*initialize number of inputs, outputs, fanin, fanout*/

  int max_delay;				/*storage for max_delay, used in line 22 of the DEE */

  int loose_inputs = 0, loose_outputs = 0, loose_bidirs = 0; 		/*number of loose wires,*/
									/*one for each type*/

  int shift = 0;  			/*used to store shift caused by removing elements*/ 
  
  static long head_symt_addr;		/* symbol table address */
  static long head_modt_addr;		/* module table address */

  long	ctime;				/* stores 

  /* missing parameter check */
    
  /* yydebug=1; */

  /**Check for correct number of parameters**/

  if (argc!=3) {
    fprintf(stderr, "Incorrect number of arguments in command line.\n"
	    "Exiting program.\n");
    fprintf(stderr, "Correct usage: parse <name of Verilog file> <name of Dee file>\n");
    exit(0);
  }

  /***** OPEN FILES *****/
  /* input file */

  if (argc>1){
    infilename=(char*)malloc(sizeof(char)*strlen(argv[1])+1);
    sprintf (infilename, "%s%s", argv[1], VERILOG_EXTENSION);
    if ((yyin=fopen(infilename,"r"))==NULL){
      sprintf (infilename, "%s%s", argv[1], DES_EXTENSION);
      if ((yyin=fopen(infilename,"r"))==NULL){	
        yyerror("unable to open inputfile");
      }
    }    
    argc--;
    argv++;
  }else{
    infilename=" ";
  }
	
  /* output file */
  if (argc>1){
    sprintf (infilename, "%s%s", argv[1], NET_EXTENSION);
    if ((yyout=fopen(infilename,"w"))==NULL){
      yyerror("could not open outputfile\n");
    }
  }
  
  yyparse();
  
  fclose(yyin);
  	
			/*List Modification*/
  for(wirelist = whead; wirelist; wirelist=wirelist->next){

	Fanout += wirelist->fanout;
	Fanin += wirelist->elem->inputs; 	
	Outputs += wirelist->elem->outputs;
	Inputs += wirelist->elem->inputs;

	if(wirelist->fanout!=0){  /*empty*/
	}
	
	else if(wirelist->fanin!=0){	/*for wire with fanin but no fanout, set as output*/

		for(wirelist->elem->conn=wirelist->elem->cinhead; wirelist->elem->conn; 
		    wirelist->elem->conn=wirelist->elem->conn->next){
		  wirelist->elem->position = OUTP;
		}						
	}	
	else{				/*for wire with no fanin or fanout, remove entirely*/
	 	HT_UNLINK(wirelist->elem, &ehead, &etail);
		if(wirelist->elem->position == INP) loose_inputs++;
		if(wirelist->elem->position == OUTP) loose_outputs++;
		if(wirelist->elem->position == BIDIR) loose_bidirs++;
		HT_UNLINK(wirelist, &whead, &wtail);
	}
  }

  PIs -= loose_inputs;			/*correction of Primary inputs, outputs, Bidirectionals, # of elems*/
  POs -= loose_outputs;
  elemcount -= (POs + loose_outputs + loose_inputs);
  BiDirectionals -= loose_bidirs;

  Fanin -= (PIs + loose_inputs);	/*correction of Fanin, Outputs, and Inputs*/
  Outputs -= (loose_outputs + loose_inputs);
  Inputs -= (loose_outputs + loose_inputs);


  wirecount = PIs + BiDirectionals;     /*Set wirecount*/

  i=0;					/*Reset indexes for modified wirelist*/
  for(wirelist = whead; wirelist; wirelist=wirelist->next){
	wirelist->elem->index = i++;
  }

			/*set first element of delay table to ZERO*/
  delay_table[0].rise = 0;
  delay_table[0].fall = 0;

	
			/*set all DFF's rise and fall delay to 1 if not specified*/
  for(elemlist = ehead; elemlist!=NULL; elemlist=elemlist->next){
    if(elemlist->logictype == E_LOGIC_GATE && elemlist->gatetype == D_FLIP_FLOP && elemlist->delay == 0){
      elemlist->sim_lev = 0;
      for(i=0; i<delay_index; i++){
	if(delay_table[i].rise == 1 && delay_table[i].fall == 1) elemlist->delay = i;
      }
      if(elemlist->delay == 0){
	delay_table[delay_index].rise = 1;
	delay_table[delay_index].fall = 1;
	elemlist->delay = delay_index;
	delay_index++;
      }
    }  
  }
			/*preparation of elemlist for calculation of sim_levels*/	
  for(elemlist = ehead; elemlist; elemlist=elemlist->next){
    if(elemlist->delay == 0 && elemlist->position != INP){
      elemlist->sim_lev = -1;		/*Initialization of "sim_lev" in element list for*/
				  	/*LevelStep() function called in following "for" loop*/
    }
  }
			/*Call LevelStep() if the element has delay or is an output element.*/
			/*An output gate or a gate with a non-zero delay is the starting    */
			/*point for sim_lev calculation.  (See LevelStep() function dscrptn)*/
  for(elemlist = ehead; elemlist; elemlist=elemlist->next){
    if(elemlist->position == OUTP || elemlist->delay != 0){
      i = LevelStep(elemlist);		/*"i" is garbage, used to start recursive LevelStep()*/
    }
  } 

			/*find max delay*/ 
  max_delay = 0;
  for(i=0; i<delay_index; i++){
	if(max_delay < delay_table[i].rise) max_delay = delay_table[i].rise;
	if(max_delay < delay_table[i].fall) max_delay = delay_table[i].fall;
  }
 
  for(wirelist = whead; wirelist; wirelist=wirelist->next){
	for(wirelist->conn=wirelist->couthead; wirelist->conn; 
		    wirelist->conn=wirelist->conn->next){
	}
  }

/*BEGIN THE HEADER INFORMATION*/
	
  	fprintf(yyout, "3.0\n");
	fprintf(yyout, "%s\n", argv[1]);
	head_symt_addr = ftell(yyout);		/*store position of SMT address location in DEE*/
			
        fprintf(yyout, "%8d\n", 0);

        head_modt_addr = ftell(yyout);		/*store position of module address location in DEE*/
			
        fprintf(yyout, "%8d\n", 0);
	fprintf(yyout, "0\n");
	fprintf(yyout, "%d 1\n", PIs);  
	fprintf(yyout, "%d 1\n", POs);  
	fprintf(yyout, "%d 0\n", BiDirectionals); 
 	fprintf(yyout, "0\n");  
	fprintf(yyout, "0\n");  
	fprintf(yyout, "%d 1\n", wirecount); 
	fprintf(yyout, "0 0\n");
	fprintf(yyout, "0\n");  
	fprintf(yyout, "%d\n", elemcount - wirecount);  
	fprintf(yyout, "0\n");  
	fprintf(yyout, "0\n");  
	fprintf(yyout, "0\n");  
	fprintf(yyout, "0\n");  
	fprintf(yyout, "0\n");  
	fprintf(yyout, "0\n");  
	fprintf(yyout, "0\n");  
	fprintf(yyout, "%d -9 0 %d\n", delay_index, max_delay); 					
	fprintf(yyout, "0\n"); 		/*# of initialization tables*/
	fprintf(yyout, "0 0\n");  
	fprintf(yyout, "0 0\n");  
	fprintf(yyout, "5 0\n");  
	fprintf(yyout, "1 0\n");  
	fprintf(yyout, "-1 -1\n");  	
	fprintf(yyout, "30031977\n");
	fprintf(yyout, "094500\n");  
	time(&ctime);          
        fprintf(yyout, "%d\n", ctime);
	fprintf(yyout, "333\n");  
	fprintf(yyout, "1 1 0 1\n");
	fprintf(yyout, "%d %d %d %d %d %d\n", elemcount, Fanin, Inputs - PIs,
		Fanout, Outputs, BiDirectionals);
	fprintf(yyout, "0 0 0\n");  
	fprintf(yyout, "0 0\n");  
	fprintf(yyout, "0 0\n");  
	fprintf(yyout, "0 0\n");  
	fprintf(yyout, "0 0\n");  
	fprintf(yyout, "0 0\n");  

/*END OF HEADER*/
/*DELAY TABLE*/
	
	for(i=0; i<delay_index; i++){
	  fprintf(yyout, "%d %d\n", delay_table[i].rise, delay_table[i].fall);
	}

/*END OF DELAY TABLE*/
/*TOPOLOGICAL INFORMATION*/
	
	  for(wirelist = whead; wirelist; wirelist=wirelist->next){
	      fprintf(yyout, "%d %d %d 0\n", wirelist->elem->position, wirelist->elem->logictype,
					     wirelist->elem->sim_lev);
	  
	      if(wirelist->elem->logictype==E_WIRE)
		fprintf(yyout, "1 128 %d %d %d %d %d %d\n", wirelist->elem->delay, wirelist->fanout, 
			wirelist->fanin, wirelist->elem->connection, wirelist->elem->outputs,
			wirelist->elem->inputs);
	      else 
   		fprintf(yyout, "%d %d 1 1 %d %d %d %d\n", wirelist->elem->gatetype, wirelist->elem->delay,
			wirelist->fanout, wirelist->elem->inputs, wirelist->elem->outputs,
			wirelist->elem->inputs);
	
              for(wirelist->conn = wirelist->couthead; wirelist->conn; 
		  wirelist->conn = wirelist->conn->next){
        	fprintf(yyout, "%d %d %d 0\n", wirelist->conn->ept->logictype,
			  wirelist->conn->ept->index, wirelist->link_bit);
	      }
	      
	      for(wirelist->elem->conn = wirelist->elem->cinhead; wirelist->elem->conn; 
		  wirelist->elem->conn = wirelist->elem->conn->next){
		fprintf(yyout, "%d %d %d 0\n", 
			  wirelist->elem->conn->wpt->elem->logictype,
			  wirelist->elem->conn->wpt->elem->index,
			  wirelist->elem->conn->wpt->link_bit);
	      }
	    }
	  
/*END OF TOPOLOGICAL INFORMATION*/

/*INITIALIZATION TABLE NOT IMPLEMENTED*/

/*SYMBOL TABLE*/

	/*Find the Symbol Table Address*/
	/* store the symbol table start address */
        sym_tb_addr = ftell(yyout);

  	fprintf(yyout, "%d 0\n", modulecount);
	fprintf(yyout, "%s 0 %d %s -1 -1 -1 0\n", argv[1], elemcount - 1, argv[1]);
	fprintf(yyout, "%s %d\n", argv[1], elemcount*2 - wirecount);
	for(wirelist = whead; wirelist!=NULL; wirelist=wirelist->next){
	  for(i=0; i<wirelist->elem->outputs; i++){
	    fprintf(yyout, "%d %d %d %s\n", wirelist->elem->index, wirelist->link_bit, 
			wirelist->elem->logictype, wirelist->elem->name);
	    if(wirelist->elem->logictype == 5){
	      if(wirelist->name != NULL)
	       	fprintf(yyout, "%d %d 1 %s\n", wirelist->elem->index, wirelist->link_bit, wirelist->name);
	      else fprintf(yyout, "%d %d 1 %s\n", wirelist->elem->index, wirelist->link_bit, 
			   wirelist->elem->name);
	    }
	  }
	}
	/* store the symbol table start address */
        mod_tb_addr = ftell(yyout);
	
	fprintf(yyout, "%s 0 0\n", argv[1]);
	
		/*find the symbol table address location in the Header and print the information*/
	fseek(yyout, head_symt_addr, 0);
        fprintf(yyout, "%8d", sym_tb_addr);

		/*find the module table address location in the Header and print the information*/		
        fseek(yyout, head_modt_addr, 0);		
        fprintf(yyout, "%8d", mod_tb_addr);

	exit(0);

/*END OF SYMBOL TABLE*/
}
/*END OF MAIN(INT ARGC, CHAR** ARGV)*/


/*******************************************************************************************************************

 FILENAME:		VERILOG.YACC

 FUNCTION NAME: 	LevelStep(ELEM *elemlist)

 CREATION DATE:		August 3, 1998
	
 AUTHOR:		Ben Foster

 PARAMETERS:	Input: 	pointer to the element list
		Output:	integer representing the simulation level of the gate which called LevelStep() 
			(Greatest sim_level of all input elements to current gate + 1)

 PURPOSE:		This function is used to cycle through all the elements in the elemlist, modifying the
			simulation level of each zero delay element.  The simulation level is used with zero
			delay elements so the MCS simulator knows in which order to evaluate them.  This 
			recursive function begins with either output elements or elements which have been 
			assigned a non-zero delay, recursively checking every element which contributes to the
			original gates inputs, returning only when it encounters a primary input, another 
			element with non-zero delay, or an element which has already been assigned a 
			simulation level value.  The value returned is the greatest of all the sim_lev values
			of all the elements which have outputs fanning into the current gate, PLUS ONE.  This 
			essentially creates an increasing depth in the circuit, starting at one for the inputs
			and increasing by one more for each level deeper into the circuit.  The sim_level value
			for all elements with a non-zero delay will always be zero.

 SIDE EFFECTS:		The purpose of the function is to calculate the value of "sim_lev" for each element.  
			Once this is done, the original elemlist->sim_lev is modified.   

 COMMENTS: 		The sim_lev value for non-zero delay elements : ZERO.  For primary input elements : ONE.

 HISTORY:		Unmodified

*******************************************************************************************************************/

int LevelStep(ELEM *elemlist){

  int temp = 0, junk, greatest = 0;
  CONNECTION *tempconn; 

  for(elemlist->conn=elemlist->cinhead; elemlist->conn; elemlist->conn=elemlist->conn->next){
 
	tempconn = elemlist->conn;	/*save the current connection*/ 

    if(elemlist->conn->wpt->elem->sim_lev == 1 && elemlist->delay == 0){	/*set temp to 2 if the previous*/
	temp = elemlist->conn->wpt->elem->sim_lev + 1;				/* element is an input element */
    }

    else if(elemlist->delay != 0){					/*throw away the value returned if*/ 
	if(elemlist->conn->wpt->elem->sim_lev != -1){			/*the element has a non-zero delay*/
	  junk = elemlist->conn->wpt->elem->sim_lev;
	}
	else junk = LevelStep(elemlist->conn->wpt->elem);		/*if the previous gate's sim_lev has*/
    }									/*not been set, set it.*/
    else{
	if(elemlist->conn->wpt->elem->sim_lev != -1){			/*if the previous gate's sim_lev HAS  */
	  temp = elemlist->conn->wpt->elem->sim_lev + 1;		/*been set, set current gate's sim_lev*/
	}								/*to previous sim_lev + 1*/
	else temp = LevelStep(elemlist->conn->wpt->elem);		/*<<==Else, set previous gate's sim_lev*/
    }
    if(temp >= greatest) greatest = temp;			/*store greatest sim_lev value for entire*/
	elemlist->conn = tempconn; 				/*input connection list of each element  */
  }
  if(elemlist->position == INP) greatest = 1;			
  elemlist->sim_lev = greatest;					/*set sim_lev to greatest value found for */
								/*current gate */
  return greatest + 1;						/*return greatest + 1 for NEXT element*/
}
/*END OF LevelStep(ELEM *elemlist)*/

/*******************************************************************************************************************

 FILENAME:		VERILOG.YACC

 FUNCTION NAME: 	ht_make (int size)

 CREATION DATE:		1992 

 AUTHOR:		Prof. Alva L. Couch

 PARAMETERS:	Input:	integer "size": specifies the number of bytes needed to allocate for the new element
		Output: returns a new element

 PURPOSE:		This function creates a new element to be used in a linked list

 SIDE EFFECTS:		None
	
 COMMENTS:		!!BETTER TO USE THE "MALLOC_TYPE(type)" MACRO DEFINED AT THE BEGINNING OF THE PROGRAM!!

 HISTORY:		Unmodified

*******************************************************************************************************************/

/* create a list element */
Htelem *ht_make (int size)
{
    Htelem *temp;
    temp = (Htelem *) malloc (size);
    temp->next = NULL;	/* set pointer to NULL here, avoid mistakes later */
    return (temp);
}
/*END OF ht_make (int size)*/

/*******************************************************************************************************************

 FILENAME:		VERILOG.YACC

 FUNCTION NAME: 	ht_tail (Htelem *elem, Htelem **head, Htelem **tail)

 CREATION DATE:		1992 

 AUTHOR:		Prof. Alva L. Couch

 PARAMETERS:	Input:	element to be added  (*elem)
			head of element list (**head)
			tail of element list (**tail)
		Output: returns the added element

 PURPOSE:		This function attaches an element to the tail of a list specified by pointers to the 
			head and tail of the list.

 SIDE EFFECTS:		The tail pointer is redirected to point at the new element, which becomes the new tail
	
 COMMENTS:		!!BETTER TO USE THE "HT_TAIL(a,h,t)" MACRO DEFINED AT THE BEGINNING OF THE PROGRAM!!

 HISTORY:		Unmodified

*******************************************************************************************************************/

/* link an element to the tail of a head-tail list */
Htelem *ht_tail (Htelem *elem, Htelem **head, Htelem **tail)
{
    if (((*head) == NULL) || ((*tail) == NULL))  {  /* list started empty */
	(*head) = elem;
	(*tail) = elem;
    } else { /* put elem after tail element */
	(*tail)->next = elem;
	(*tail) = elem;
    }
    elem->next = NULL;
    return (elem);
} /* end of ht_tail */

/*******************************************************************************************************************

 FILENAME:		VERILOG.YACC

 FUNCTION NAME: 	ht_head (Htelem *elem, Htelem **head, Htelem **tail)

 CREATION DATE:		1992 

 AUTHOR:		Prof. Alva L. Couch

 PARAMETERS:	Input:	element to be added  (*elem)
			head of element list (**head)
			tail of element list (**tail)
		Output: returns the added element

 PURPOSE:		This function attaches an element to the head of a list specified by pointers to the 
			head and tail of the list.

 SIDE EFFECTS:		The head pointer is redirected to point at the new element, which becomes the new head
	
 COMMENTS:		!!BETTER TO USE THE "HT_HEAD(a,h,t)" MACRO DEFINED AT THE BEGINNING OF THE PROGRAM!!

 HISTORY:		Unmodified

*******************************************************************************************************************/
    
/* link an element to the head of a head-tail list */
Htelem *ht_head (Htelem *elem, Htelem **head, Htelem **tail)
{
    if (((*head) == NULL) || ((*tail) == NULL))  {  /* list started empty */
	(*head) = elem;
	(*tail) = elem;
	elem->next = NULL;
    } else { /* put elem before head element */
	elem->next = (*head);
	(*head) = elem;
    }
    return (elem);
} /* end of ht_head */

/*******************************************************************************************************************

 FILENAME:		VERILOG.YACC

 FUNCTION NAME: 	ht_unlink (Htelem *elem, Htelem **head, Htelem **tail)

 CREATION DATE:		1992 

 AUTHOR:		Prof. Alva L. Couch

 PARAMETERS:	Input:	element to be deleted (*elem)
			head of element list (**head)
			tail of element list (**tail)
		Output: returns the added element

 PURPOSE:		This function deletes an element from the list containing it.  This list is defined by
			by the pointers to the list's head and tail.

 SIDE EFFECTS:		The predecessor in the list's pointer to the next element (the element being unlinked)
			is redirected.
	
 COMMENTS:		!!BETTER TO USE THE "HT_UNLINK(a,h,t)" MACRO DEFINED AT THE BEGINNING OF THE PROGRAM!!

 HISTORY:		Unmodified

*******************************************************************************************************************/
    
/* unlink a segment from the master list */
Htelem *ht_unlink(Htelem *elem, Htelem **head, Htelem **tail) {
    Htelem *pred;

    if (elem == NULL) ;					/* nothing */

    else if (((*head) == NULL) || ((*tail) == NULL)) {	/* empty list */
        (*head) = NULL;
        (*tail) = NULL;
    } 

    else if (((*head) == elem) && ((*tail) == elem)) {	/* last elem  */
        (*head) = NULL;
        (*tail) = NULL;
    } 

    else if ((*head) == elem) {				/* off of top */
        (*head) = (*head)->next;
    } 

    else { 	   /* find the predecessor to segment, link around it */
        for (pred=(*head); 
	     (pred->next!=elem) && (pred->next!=NULL); 
	     pred=pred->next) ;
        if (pred->next == elem) {
	    pred->next = pred->next->next;
            if (elem == (*tail)) 
		(*tail) = pred; 
	}
    } 
    return (elem);
}  /* end: ht_unlink */

